import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.modules import DenseEmbedding, OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes, PoolingWeightedNodes
from kgcnn.utils.models import update_model_kwargs, generate_embedding

# 'Semi-Supervised Classification with Graph Convolutional Networks'
# by Thomas N. Kipf, Max Welling
# https://arxiv.org/abs/1609.02907
# https://github.com/tkipf/gcn

model_default = {'name': "GCN",
                 'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None, 1), 'name': "edge_attributes", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                 'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                     "edge": {"input_dim": 10, "output_dim": 64}},
                 'gcn_args': {"units": 100, "use_bias": True, "activation": 'relu', "pooling_method": 'sum',
                              "is_sorted": False, "has_unconnected": True},
                 'depth': 3, 'verbose': 10,
                 'output_embedding': 'graph',
                 'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
                                "activation": ['relu', 'relu', 'sigmoid']}
                 }


@update_model_kwargs(model_default)
def make_model(inputs=None,
               input_embedding=None,
               depth=None,
               gcn_args=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make GCN graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        gcn_args (dict): Dictionary of layer arguments unpacked in `GCN` convolutional layer.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """
    input_node_shape = inputs[0]['shape']
    input_edge_shape = inputs[1]['shape']

    if input_edge_shape[-1] != 1:
        raise ValueError("No edge features available for GCN, only edge weights of pre-scaled adjacency matrix, \
                         must be shape (batch, None, 1), but got (without batch-dimension): ", input_edge_shape)

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])

    # Embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(input_node_shape) < 2)(node_input)
    ed = OptionalInputEmbedding(**input_embedding['edge'],
                                use_embedding=len(input_edge_shape) < 2)(edge_input)
    edi = edge_index_input

    # Model
    n = DenseEmbedding(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
    for i in range(0, depth):
        n = GCN(**gcn_args)([n, ed, edi])

    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes()(n)  # will return tensor
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = n
        out = GraphMLP(**output_mlp)(out)
        # no ragged for distribution supported atm
        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported graph embedding for `GCN`")

    model = ks.models.Model(inputs=[node_input, edge_input, edge_index_input], outputs=out)
    return model


model_default_weighted = {'name': "GCN_weighted",
                          'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                                     {'shape': (None, 1), 'name': "edge_attributes", 'dtype': 'float32', 'ragged': True},
                                     {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True},
                                     {'shape': (None, 1), 'name': "node_weights", 'dtype': 'float32', 'ragged': True}],
                          'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                              "edge": {"input_dim": 10, "output_dim": 64}},
                          'output_embedding': 'graph',
                          'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
                                         "activation": ['relu', 'relu', 'sigmoid']},
                          'gcn_args': {"units": 100, "use_bias": True, "activation": 'relu', "pooling_method": 'sum'},
                          'depth': 3, 'verbose': 1
                          }


@update_model_kwargs(model_default_weighted)
def make_model_weighted(inputs=None,
                        input_embedding=None,
                        output_embedding=None,
                        output_mlp=None,
                        depth=None,
                        gcn_args=None,
                        name=None,
                        verbose=None):
    """Make GCN model with edge weights."""

    input_node_shape = inputs[0]['shape']
    input_edge_shape = inputs[1]['shape']

    if input_edge_shape[-1] != 1:
        raise ValueError("No edge features available for GCN, only edge weights of pre-scaled adjacency matrix, \
                         must be shape (batch, None, 1), but got (without batch-dimension): ", input_edge_shape)

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    node_weights_input = ks.layers.Input(**inputs[3])

    # Embedding, if no feature dimension
    n = generate_embedding(node_input, input_node_shape, input_embedding['node'])
    ed = generate_embedding(edge_input, input_edge_shape, input_embedding['edge'])
    edi = edge_index_input
    nw = node_weights_input

    # Model
    n = DenseEmbedding(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
    for i in range(0, depth):
        n = GCN(**gcn_args)([n, ed, edi])

    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingWeightedNodes()([n, nw])  # will return tensor
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = n
        out = GraphMLP(**output_mlp)(out)
        # no ragged for distribution supported atm
        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported graph embedding for `GCN`")

    model = ks.models.Model(inputs=[node_input, edge_input, edge_index_input, node_weights_input], outputs=out)
    return model



import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.conv.attention import AttentionHeadGATV2
from kgcnn.layers.modules import LazyConcatenate, DenseEmbedding, LazyAverage, ActivationEmbedding, \
    OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.utils.models import update_model_kwargs

# Graph Attention Networks by Veličković et al. (2018)
# https://arxiv.org/abs/1710.10903
# Improved by
# How Attentive are Graph Attention Networks?
# by Brody et al. (2021)

model_default = {'name': "GATv2",
                 'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None,), 'name': "edge_attributes", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                 'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                     "edge": {"input_dim": 5, "output_dim": 64}},
                 'attention_args': {"units": 32, "use_final_activation": False, "use_edge_features": True,
                                    "has_self_loops": True, "activation": "kgcnn>leaky_relu", "use_bias": True},
                 'pooling_nodes_args': {'pooling_method': 'mean'},
                 'depth': 3, 'attention_heads_num': 5,
                 'attention_heads_concat': False, 'verbose': 10,
                 'output_embedding': 'graph',
                 'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
                                "activation": ['relu', 'relu', 'sigmoid']}
                 }


@update_model_kwargs(model_default)
def make_model(inputs=None,
               input_embedding=None,
               attention_args=None,
               pooling_nodes_args=None,
               depth=None,
               attention_heads_num=None,
               attention_heads_concat=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make GATv2 graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        attention_args (dict): Dictionary of layer arguments unpacked in `AttentionHeadGATV2` layer.
        pooling_nodes_args (dict): Dictionary of layer arguments unpacked in `PoolingNodes` layer.
        depth (int): Number of graph embedding units or depth of the network.
        attention_heads_num (int): Number of attention heads to use.
        attention_heads_concat (bool): Whether to concat attention heads. Otherwise average heads.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])

    # Embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    ed = OptionalInputEmbedding(**input_embedding['edge'],
                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
    edi = edge_index_input

    # Model
    nk = DenseEmbedding(units=attention_args["units"], activation="linear")(n)
    for i in range(0, depth):
        heads = [AttentionHeadGATV2(**attention_args)([nk, ed, edi]) for _ in range(attention_heads_num)]
        if attention_heads_concat:
            nk = LazyConcatenate(axis=-1)(heads)
        else:
            nk = LazyAverage()(heads)
            nk = ActivationEmbedding(activation=attention_args["activation"])(nk)
    n = nk

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_nodes_args)(n)
        main_output = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)(n)
        main_output = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported graph embedding for `GATv2`")

    # Define model output
    model = tf.keras.models.Model(inputs=[node_input, edge_input, edge_index_input], outputs=main_output)
    return model



import logging
import numpy as np
import tensorflow as tf
import networkx as nx
import pandas as pd
import os
import re

from kgcnn.utils.adj import get_angle_indices, coordinates_to_distancematrix, invert_distance, \
    define_adjacency_from_distance, sort_edge_indices, get_angle, add_edges_reverse_indices, \
    rescale_edge_weights_degree_sym, add_self_loops_to_edge_indices, compute_reverse_edges_index_map
from kgcnn.utils.data import save_pickle_file, load_pickle_file, ragged_tensor_from_nested_numpy

logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class GraphDict(dict):
    r"""Dictionary container to store graph information in tensor-form. At the moment only numpy-arrays are supported.
    The naming convention is not restricted. The class is supposed to be handled just as a python dictionary.
    In addition, :obj:`assign_property` and :obj:`obtain_property` handles `None` values and cast into tensor format,
    when assigning a named value.
    Graph operations that modify edges or sort indices are methods of this class. Note that the graph-tensors name
    must follow a standard-convention or be provided to member functions (see documentation of class-methods).

    .. code-block:: python

        import numpy as np
        from kgcnn.data.base import GraphDict
        g = GraphDict({"edge_indices": np.array([[1, 0], [0, 1]]), "edge_labels": np.array([[-1], [1]])})
        g.add_edge_self_loops().sort_edge_indices()
        print(g)
    """

    # Implementation details: Inherits from python-dict at the moment but can be changed if this causes problems,
    # alternatives would be: collections.UserDict or collections.abc.MutableMapping

    def __init__(self, sub_dict: dict = None):
        r"""Initialize a new :obj:`GraphDict` instance.

        Args:
            sub_dict: Dictionary or key-value pair of numpy arrays.
        """
        self._tensor_conversion = np.array
        if sub_dict is None:
            sub_dict = {}
        elif isinstance(sub_dict, (dict, list)):
            in_dict = dict(sub_dict)
            sub_dict = {key: self._tensor_conversion(value) for key, value in in_dict.items()}
        elif isinstance(sub_dict, GraphDict):
            sub_dict = {key: self._tensor_conversion(value) for key, value in sub_dict.items()}
        super(GraphDict, self).__init__(sub_dict)

    def to_dict(self) -> dict:
        """Returns a python-dictionary of self. Does not copy values.

        Returns:
            dict: Dictionary of graph tensor objects.
        """
        return {key: value for key, value in self.items()}

    def from_networkx(self, graph,
                      node_number: str = "node_number",
                      edge_indices: str = "edge_indices",
                      node_attributes: str = None,
                      edge_attributes: str = None,
                      node_labels: str = None):
        r"""Convert a networkx graph instance into a dictionary of graph-tensors. The networkx graph is always converted
        into integer node labels. The former node IDs can be hold in :obj:`node_labels`. Furthermore, node or edge
        data can be cast into attributes via :obj:`node_attributes` and :obj:`edge_attributes`.

        Args:
            graph (nx.Graph): A networkx graph instance to convert.
            node_number (str): The name that the node numbers are assigned to. Default is "node_number".
            edge_indices (str): The name that the edge indices are assigned to. Default is "edge_indices".
            node_attributes (str, list): Name of node attributes to add from node data. Can also be a list of names.
                Default is None.
            edge_attributes (str, list): Name of edge attributes to add from edge data. Can also be a list of names.
                Default is None.
            node_labels (str): Name of the labels of nodes to store former node IDs into. Default is None.

        Returns:
            self.
        """
        assert node_labels is None or isinstance(node_labels, str), "Please provide name of node labels or `None`"
        graph_int = nx.convert_node_labels_to_integers(graph, label_attribute=node_labels)
        graph_size = len(graph_int)

        def _attr_to_list(attr):
            if attr is None:
                attr = []
            elif isinstance(attr, str):
                attr = [attr]
            if not isinstance(attr, list):
                raise TypeError("Attribute name is neither list nor string.")
            return attr

        # Loop over nodes in graph.
        node_attr = _attr_to_list(node_attributes)
        if node_labels is not None:
            node_attr += [node_labels]
        node_attr_dict = {x: [None]*graph_size for x in node_attr}
        nodes_id = []
        for i, x in enumerate(graph_int.nodes.data()):
            nodes_id.append(x[0])
            for d in node_attr:
                if d not in x[1]:
                    raise KeyError("Node does not have property %s" % d)
                node_attr_dict[d][i] = x[1][d]

        edge_id = []
        edges_attr = _attr_to_list(edge_attributes)
        edges_attr_dict = {x: [None]*graph_size for x in edges_attr}
        for i, x in enumerate(graph_int.edges.data()):
            edge_id.append(x[:2])
            for d in edges_attr:
                if d not in x[2]:
                    raise KeyError("Edge does not have property %s" % d)
                edges_attr_dict[d][i] = x[2][d]

        # Storing graph tensors in self.
        self.assign_property(node_number, self._tensor_conversion(nodes_id))
        self.assign_property(edge_indices, self._tensor_conversion(edge_id))
        for key, value in node_attr_dict.items():
            self.assign_property(key, self._tensor_conversion(value))
        for key, value in edges_attr_dict.items():
            self.assign_property(key, self._tensor_conversion(value))
        return self

    def to_networkx(self, edge_indices="edge_indices"):
        """Function draft to make a networkx graph. No attributes or data is supported at the moment.

        Args:
            edge_indices (str): Name of edge index tensors to make graph with. Default is "edge_indices".

        Returns:
            nx.DiGraph: Directed networkx graph instance.
        """
        graph = nx.DiGraph()
        graph.add_edges_from(self.obtain_property(edge_indices))
        return graph

    def assign_property(self, key: str, value):
        r"""Add a named property as key, value pair to self. If the value is `None`, nothing is done.
        Similar to assign-item default method :obj:`__setitem__`, but ignores `None` values and casts to tensor.

        Args:
            key (str): Name of the graph tensor to add to self.
            value: Array or tensor value to add. Can also be None.

        Returns:
            None.
        """
        if value is not None:
            self.update({key: self._tensor_conversion(value)})

    def obtain_property(self, key: str):
        """Get tensor item by name. If key is not found, `None` is returned.

        Args:
            key (str): Name of the key to get value for.

        Returns:
            self[key].
        """
        if key in self:
            return self[key]

    def _find_graph_properties(self, name_props: str) -> list:
        r"""Search for properties in self. This includes a list of possible names or a pattern-matching of a single
        string.

        Args:
            name_props (str, list): Pattern matching string or list of strings to search for

        Returns:
            list: List of names in self that match :obj:`name_props`.
        """
        if name_props is None:
            return []
        elif isinstance(name_props, str):
            match_props = []
            for x in self:
                if re.match(name_props, x):
                    if re.match(name_props, x).group() == x:
                        match_props.append(x)
            return match_props
        elif isinstance(name_props, (list, tuple)):
            return [x for x in name_props if x in self]
        raise TypeError("Can not find keys of properties for input type %s" % name_props)

    def _assert_has_key(self, key: str):
        """Check if the property is found in self.

        Args:
            key (str): Name of property that must be defined.

        Returns:
            None.
        """
        if key not in self or self[key] is None:
            raise ValueError("Can not operate on %s, as it is not found." % key)

    def _operate_on_edges(self, operation,
                          edge_indices: str = "edge_indices",
                          edge_attributes: str = "^edge_.*",
                          **kwargs):
        r"""General wrapper to run a certain function on an array of edge-indices and all edge related properties
        in the dictionary of self.
        The name of the key for indices must be defined in :obj:`edge_indices`. Related value or attribute tensors
        are searched by :obj:`edge_attributes`, since they must behave accordingly.

        Args:
              operation (callable): Function to apply to a list of all edge arrays.
                First entry is assured to be indices.
              edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
              edge_attributes (str): Name of related edge values or attributes.
                This can be a match-string or list of names. Default is "^edge_.*".
              kwargs: Kwargs for operation function call.

        Returns:
            self
        """
        self._assert_has_key(edge_indices)
        # Determine all linked edge attributes, that are not None.
        edge_linked = self._find_graph_properties(edge_attributes)
        # Edge indices is always at first position!
        edge_linked = [edge_indices] + [x for x in edge_linked if x != edge_indices]
        no_nan_edge_prop = [x for x in edge_linked if self[x] is not None]
        non_nan_edge = [self[x] for x in no_nan_edge_prop]
        new_edges = operation(*non_nan_edge, **kwargs)
        # If dataset only has edge indices, operation is expected to only return array not list!
        # This restricts the type of operation used with this method.
        if len(no_nan_edge_prop) == 1:
            new_edges = [new_edges]
        # Set all new edge attributes.
        for i, at in enumerate(no_nan_edge_prop):
            self[at] = new_edges[i]
        return self

    def set_edge_indices_reverse(self, edge_indices: str = "edge_indices",
                                 edge_indices_reverse: str = "edge_indices_reverse"):
        r"""Computes the index map of the reverse edge for each of the edges, if available. This can be used by a model
        to directly select the corresponding edge of :math:`(j, i)` which is :math:`(i, j)`.
        Does not affect other edge-properties, only creates a map on edge indices. Edges that do not have a reverse
        pair get a `nan` as map index. If there are multiple edges, the first encounter is assigned.

        .. warning::
            Reverse maps are not recomputed if you use e.g. :obj:`sort_edge_indices` or redefine edges.

        Args:
            edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
            edge_indices_reverse (str): Name of reverse indices to store output. Default is "edge_indices_reverse"

        Returns:
            self
        """
        self._assert_has_key(edge_indices)
        self[edge_indices_reverse] = np.expand_dims(
            compute_reverse_edges_index_map(self[edge_indices]), axis=-1)
        return self

    def make_undirected_edges(self, edge_indices: str = "edge_indices",
                              edge_attributes: str = "^edge_.*",
                              remove_duplicates: bool = True,
                              sort_indices: bool = True):
        r"""Add edges :math:`(j, i)` for :math:`(i, j)` if there is no edge :math:`(j, i)`.
        With :obj:`remove_duplicates` an edge can be added even though there is already and edge at :math:`(j, i)`.
        For other edge tensors, like the attributes or labels, the values of edge :math:`(i, j)` is added in place.
        Requires that :obj:`edge_indices` property is assigned.

        Args:
            edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
            edge_attributes (str): Name of related edge values or attributes.
                This can be a match-string or list of names. Default is "^edge_.*".
            remove_duplicates (bool): Whether to remove duplicates within the new edges. Default is True.
            sort_indices (bool): Sort indices after adding edges. Default is True.

        Returns:
            self
        """
        self._operate_on_edges(add_edges_reverse_indices, edge_indices=edge_indices,
                               edge_attributes=edge_attributes,
                               remove_duplicates=remove_duplicates, sort_indices=sort_indices)
        return self

    def add_edge_self_loops(self, edge_indices: str = "edge_indices",
                            edge_attributes: str = "^edge_.*",
                            remove_duplicates: bool = True,
                            sort_indices: bool = True,
                            fill_value: int = 0):
        r"""Add self loops to the each graph property. The function expects the property :obj:`edge_indices`
        to be defined. By default the edges are also sorted after adding the self-loops.
        All other edge properties are filled with :obj:`fill_value`.

        Args:
            edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
            edge_attributes (str): Name of related edge values or attributes.
                This can be a match-string or list of names. Default is "^edge_.*".
            remove_duplicates (bool): Whether to remove duplicates. Default is True.
            sort_indices (bool): To sort indices after adding self-loops. Default is True.
            fill_value (in): The fill_value for all other edge properties.

        Returns:
            self
        """
        self._assert_has_key(edge_indices)
        self._operate_on_edges(add_self_loops_to_edge_indices, edge_indices=edge_indices,
                               edge_attributes=edge_attributes,
                               remove_duplicates=remove_duplicates, sort_indices=sort_indices, fill_value=fill_value)
        return self

    def sort_edge_indices(self, edge_indices: str = "edge_indices",
                          edge_attributes: str = "^edge_.*"):
        r"""Sort edge indices and all edge-related properties. The index list is sorted for the first entry.

        Args:
            edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
            edge_attributes (str): Name of related edge values or attributes.
                This can be a match-string or list of names. Default is "^edge_.*".

        Returns:
            self
        """
        self._assert_has_key(edge_indices)
        self._operate_on_edges(sort_edge_indices, edge_indices=edge_indices,
                               edge_attributes=edge_attributes)
        return self

    def normalize_edge_weights_sym(self, edge_indices: str = "edge_indices",
                                   edge_weights: str = "edge_weights"):
        r"""Normalize :obj:`edge_weights` using the node degree of each row or column of the adjacency matrix.
        Normalize edge weights as :math:`\tilde{e}_{i,j} = d_{i,i}^{-0.5} \, e_{i,j} \, d_{j,j}^{-0.5}`.
        The node degree is defined as :math:`D_{i,i} = \sum_{j} A_{i, j}`. Requires the property :obj:`edge_indices`.
        Does not affect other edge-properties and only sets :obj:`edge_weights`.

        Args:
            edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
            edge_weights (str): Name of edge weights indices to set in dictionary. Default is "edge_weights".

        Returns:
            self
        """
        self._assert_has_key(edge_indices)
        # If weights is not set, initialize with weight one.
        if edge_weights not in self or self.obtain_property(edge_weights) is None:
            self[edge_weights] = np.ones((len(self.obtain_property(edge_indices)), 1))
        self[edge_weights] = rescale_edge_weights_degree_sym(
            self[edge_indices], self[edge_weights])
        return self

    def set_range_from_edges(self, edge_indices: str = "edge_indices",
                             range_indices: str = "range_indices",
                             node_coordinates: str = "node_coordinates",
                             range_attributes: str = "range_attributes",
                             do_invert_distance: bool = False):
        r"""Assigns range indices and attributes (distance) from the definition of edge indices. This operations
        requires the attributes :obj:`node_coordinates` and :obj:`edge_indices` to be set. That also means that
        :obj:`range_indices` will be equal to :obj:`edge_indices`.

        Args:
            edge_indices (str): Name of indices in dictionary. Default is "edge_indices".
            range_indices (str): Name of range indices to set in dictionary. Default is "range_indices".
            node_coordinates (str): Name of coordinates in dictionary. Default is "node_coordinates".
            range_attributes (str): Name of range distance to set in dictionary. Default is "range_attributes".
            do_invert_distance (bool): Invert distance when computing  :obj:`range_attributes`. Default is False.

        Returns:
            self
        """
        self._assert_has_key(edge_indices)
        self[range_indices] = np.array(self[edge_indices], dtype="int")  # We make a copy.

        if node_coordinates not in self or self[node_coordinates] is None:
            module_logger.error("Coordinates are not set. Can not calculate range values.")
            return self
        xyz = self[node_coordinates]
        idx = self[range_indices]
        dist = np.sqrt(np.sum(np.square(xyz[idx[:, 0]] - xyz[idx[:, 1]]), axis=-1, keepdims=True))
        if do_invert_distance:
            dist = invert_distance(dist)
        self[range_attributes] = dist
        return self

    def set_range(self, range_indices: str = "range_indices",
                  node_coordinates: str = "node_coordinates",
                  range_attributes: str = "range_attributes",
                  max_distance: float = 4.0, max_neighbours: int = 15,
                  do_invert_distance: bool = False, self_loops: bool = False, exclusive: bool = True):
        r"""Define range in euclidean space for interaction or edge-like connections. The number of connection is
        determines based on a cutoff radius and a maximum number of neighbours or both.
        Requires :obj:`node_coordinates` to be set. The distance is stored in :obj:`range_attributes`.

        Args:
            range_indices (str): Name of range indices to set in dictionary. Default is "range_indices".
            node_coordinates (str): Name of coordinates in dictionary. Default is "node_coordinates".
            range_attributes (str): Name of range distance to set in dictionary. Default is "range_attributes".
            max_distance (float): Maximum distance or cutoff radius for connections. Default is 4.0.
            max_neighbours (int): Maximum number of allowed neighbours for a node. Default is 15.
            do_invert_distance (bool): Whether to invert the the distance. Default is False.
            self_loops (bool): If also self-interactions with distance 0 should be considered. Default is False.
            exclusive (bool): Whether both max_neighbours and max_distance must be fulfilled. Default is True.

        Returns:
            self
        """
        if node_coordinates not in self or self[node_coordinates] is None:
            module_logger.error("Coordinates are not set. Can not compute range.")
            return self
        # Compute distance matrix here. May be problematic for too large graphs.
        dist = coordinates_to_distancematrix(self[node_coordinates])
        cons, indices = define_adjacency_from_distance(dist, max_distance=max_distance,
                                                       max_neighbours=max_neighbours,
                                                       exclusive=exclusive, self_loops=self_loops)
        mask = np.array(cons, dtype="bool")
        dist_masked = dist[mask]
        if do_invert_distance:
            dist_masked = invert_distance(dist_masked)
        # Need one feature dimension.
        if len(dist_masked.shape) <= 1:
            dist_masked = np.expand_dims(dist_masked, axis=-1)
        # Assign attributes to instance.
        self[range_attributes] = dist_masked
        self[range_indices] = indices
        return self

    def set_angle(self, range_indices: str = "range_indices",
                  node_coordinates: str = "node_coordinates",
                  angle_indices: str = "angle_indices",
                  angle_indices_nodes: str = "angle_indices_nodes",
                  angle_attributes: str = "angle_attributes",
                  allow_multi_edges: bool = False,
                  compute_angles: bool = True):
        r"""Find possible angles between geometric range connections defined by :obj:`range_indices`.
        Which edges form angles is stored in :obj:`angle_indices`.
        One can also change :obj:`range_indices` to `edge_indices` to compute angles between edges instead
        of range connections.

        .. warning::
            Angles are not recomputed if you use :obj:`set_range` or redefine range or edges.

        Args:
            range_indices (str): Name of range indices in dictionary. Default is "range_indices".
            node_coordinates (str): Name of coordinates in dictionary. Default is "node_coordinates".
            angle_indices (str): Name of angle (edge) indices to set in dictionary. Default is "angle_indices".
            angle_indices_nodes (str): Name of angle (node) indices to set in dictionary.
                Index triplets referring to nodes. Default is "angle_indices_nodes".
            angle_attributes (str): Name of angle values to set in dictionary. Default is "angle_attributes".
            allow_multi_edges (bool): Whether to allow angles between 'i<-j<-i', which gives 0 degree angle, if the
                the nodes are unique. Default is False.
            compute_angles (bool): Whether to also compute angles.

        Returns:
            self
        """
        self._assert_has_key(range_indices)
        # Compute angles
        _, a_triples, a_indices = get_angle_indices(self[range_indices],
                                                    allow_multi_edges=allow_multi_edges)
        self[angle_indices] = a_indices
        self[angle_indices_nodes] = a_triples
        # Also compute angles
        if compute_angles:
            if node_coordinates not in self or self[node_coordinates] is None:
                module_logger.error("Coordinates are not set. Can not compute angle values.")
                return self
            self[angle_attributes] = get_angle(self[node_coordinates], a_triples)
        return self


GraphNumpyContainer = GraphDict


class MemoryGraphList:
    r"""Class to store a list of graph dictionaries in memory. Contains a python list as property :obj:`_list`.
    The graph properties are defined by tensor-like numpy arrays for indices, attributes, labels, symbol etc. .
    They are distributed from of a list of numpy arrays to each graph in the list via :obj:`assign_property`.

    A list of numpy arrays can also be passed to class instances that have a reserved prefix, which is essentially
    a shortcut to :obj:`assign_property` and must match the length of the list.
    Prefix are `node_`, `edge_`, `graph_` `range_` and `angle_` for their node, edge and graph properties, respectively.
    The range-attributes and range-indices are just like edge-indices but refer to a geometric annotation.
    This allows to have geometric range-connections and topological edges separately.
    The prefix 'range' is synonym for a geometric edge.

    .. code-block:: python

        import numpy as np
        from kgcnn.data.base import MemoryGraphList
        data = MemoryGraphList()
        data.empty(1)
        data.assign_property("edge_indices", [np.array([[0, 1], [1, 0]])])
        data.assign_property("node_labels", [np.array([[0], [1]])])
        print(data.obtain_property("edge_indices"))
        data.assign_property("node_coordinates", [np.array([[1, 0, 0], [0, 1, 0], [0, 2, 0], [0, 3, 0]])])
        print(data.obtain_property("node_coordinates"))
        data.map_list("set_range", max_distance=1.5, max_neighbours=10, self_loops=False)
        # Shortcut of reserved attributes
        print(data.range_indices, data.range_attributes)
        print(data[0])
    """

    def __init__(self):
        r"""Initialize an empty :obj:`MemoryGraphList` instance. If you want to expand the list or
        namespace of accepted reserved graph prefix identifier, you can expand :obj:`_reserved_graph_property_prefix`.

        """
        self._list = []
        self._reserved_graph_property_prefix = ["node_", "edge_", "graph_", "range_", "angle_"]
        self.logger = module_logger

    def assign_property(self, key, value):
        if value is None:
            # We could also here remove the key from all graphs.
            return self
        if not isinstance(value, list):
            raise TypeError("Expected type 'list' to assign graph properties.")
        if len(self._list) == 0:
            self.empty(len(value))
        if len(self._list) != len(value):
            raise ValueError("Can only store graph attributes from list with same length.")
        for i, x in enumerate(value):
            self._list[i].assign_property(key, x)
        return self

    def obtain_property(self, key):
        r"""Returns a list with the values of all the graphs defined for the string property name `key`. If none of
        the graphs in the list have this property, returns None.

        Args:
            key (str): The string name of the property to be retrieved for all the graphs contained in this list
        """
        # "_list" is a list of GraphNumpyContainers, which means "prop_list" here will be a list of all the property
        # values for teach of the graphs which make up this list.
        prop_list = [x.obtain_property(key) for x in self._list]

        # If a certain string property is not set for a GraphNumpyContainer, it will still return None. Here we check:
        # If all the items for our given property name are None then we know that this property is generally not
        # defined for any of the graphs in the list.
        if all([x is None for x in prop_list]):
            self.logger.warning("Property %s is not set on any graph." % key)
            return None

        return prop_list

    def __setattr__(self, key, value):
        """Setter that intercepts reserved attributes and stores them in the list of graph containers."""
        if not hasattr(self, "_reserved_graph_property_prefix") or not hasattr(self, "_list"):
            return super(MemoryGraphList, self).__setattr__(key, value)
        if any([x == key[:len(x)] for x in self._reserved_graph_property_prefix]):
            self.assign_property(key, value)
        else:
            return super(MemoryGraphList, self).__setattr__(key, value)

    def __getattribute__(self, key):
        """Getter that retrieves a list of properties from graph containers."""
        if key in ["_reserved_graph_property_prefix", "_list"]:
            return super(MemoryGraphList, self).__getattribute__(key)
        if any([x == key[:len(x)] for x in self._reserved_graph_property_prefix]):
            return self.obtain_property(key)
        else:
            return super().__getattribute__(key)

    def __len__(self):
        """Return the current length of this instance."""
        return len(self._list)

    def __getitem__(self, item):
        # Does not make a copy of the data, as a python list does.
        if isinstance(item, int):
            return self._list[item]
        new_list = MemoryGraphList()
        if isinstance(item, slice):
            return new_list._set_internal_list(self._list[item])
        if isinstance(item, list):
            return new_list._set_internal_list([self._list[int(i)] for i in item])
        if isinstance(item, np.ndarray):
            return new_list._set_internal_list([self._list[int(i)] for i in item])
        raise TypeError("Unsupported type for MemoryGraphList items.")

    def _set_internal_list(self, value: list):
        if not isinstance(value, list):
            raise TypeError("Must set list for MemoryGraphList.")
        self._list = value
        return self

    def __setitem__(self, key, value):
        if not isinstance(value, GraphNumpyContainer):
            raise TypeError("Require a GraphNumpyContainer as list item.")
        self._list[key] = value

    def clear(self):
        self._list = []

    def empty(self, length: int):
        if length is None:
            return self
        if length < 0:
            raise ValueError("Length of empty list must be >=0.")
        self._list = [GraphNumpyContainer() for _ in range(length)]
        return self

    @property
    def length(self):
        return len(self._list)

    @length.setter
    def length(self, value: int):
        raise ValueError("Can not set length. Please use 'empty()' to initialize an empty list.")

    def _to_tensor(self, item, make_copy=True):
        if not make_copy:
            self.logger.warning("At the moment always a copy is made for tensor().")
        props = self.obtain_property(item["name"])  # Will be list.
        is_ragged = item["ragged"] if "ragged" in item else False
        if is_ragged:
            return ragged_tensor_from_nested_numpy(props)
        else:
            return tf.constant(np.array(props))

    def tensor(self, items, make_copy=True):
        if isinstance(items, dict):
            return self._to_tensor(items, make_copy=make_copy)
        elif isinstance(items, (tuple, list)):
            return [self._to_tensor(x, make_copy=make_copy) for x in items]
        else:
            raise TypeError("Wrong type, expected e.g. [{'name': 'edge_indices', 'ragged': True}, {...}, ...]")

    def map_list(self, fun, **kwargs):
        for x in self._list:
            getattr(x, fun)(**kwargs)
        return self

    def clean(self, inputs: list):
        """Given a list of property names, this method removes all elements from the internal list of
        GraphNumpyContainers, which do not define at least one of those properties. Aka only those graphs remain in
        the list which definitely define all of the properties.

        Args:
            inputs (list): A list of strings, where each string is supposed to be a property name, which the graphs
                in this list may possess.
        """
        invalid_graphs = []
        for item in inputs:
            if isinstance(item, dict):
                item_name = item["name"]
            else:
                item_name = item
            props = self.obtain_property(item_name)
            if props is None:
                self.logger.warning("Can not clean property %s as it was not assigned to any graph." % item)
                continue
            for i, x in enumerate(props):
                # If property is neither list nor np.array
                if x is None or not hasattr(x, "__getitem__"):
                    self.logger.info("Property %s is not defined for graph %s." % (item_name, i))
                    invalid_graphs.append(i)
                elif not isinstance(x, np.ndarray):
                    self.logger.info("Property %s is not a numpy array for graph %s." % (item_name, i))
                    invalid_graphs.append(i)
                elif len(x.shape) > 0:
                    if len(x) <= 0:
                        self.logger.info("Property %s is empty list for graph %s." % (item_name, i))
                        invalid_graphs.append(i)
        invalid_graphs = np.unique(np.array(invalid_graphs, dtype="int"))
        invalid_graphs = np.flip(invalid_graphs)  # Need descending order
        if len(invalid_graphs) > 0:
            self.logger.warning("Found invalid graphs for properties. Removing graphs %s." % invalid_graphs)
        else:
            self.logger.info("No invalid graphs for assigned properties found.")
        for i in invalid_graphs:
            self._list.pop(int(i))
        return self


class MemoryGraphDataset(MemoryGraphList):
    r"""Dataset class for lists of graph tensor properties that can be cast into the :obj:`tf.RaggedTensor` class.
    The graph list is expected to only store numpy arrays in place of the each node or edge information!

    .. note::
        Each graph attribute is expected to be a python list or iterable object containing numpy arrays.
        For example, the special attribute of :obj:`edge_indices` is expected to be a list of arrays of
        shape `(Num_edges, 2)` with the indices of node connections.
        The node attributes in :obj:`node_attributes` are numpy arrays of shape `(Num_nodes, Num_features)`.

    The Memory Dataset class inherits from :obj:`MemoryGraphList` and has further information
    about a location on disk, i.e. a file directory and a file name as well as a name of the dataset.

    .. code-block:: python

        from kgcnn.data.base import MemoryGraphDataset
        dataset = MemoryGraphDataset(data_directory="", dataset_name="Example")
        dataset.assign_property("edge_indices", [np.array([[1, 0], [0, 1]])])
        dataset.assign_property("edge_labels", [np.array([[0], [1]])])

    The file directory and file name are used in child classes and in :obj:`save` and :obj:`load`.
    """

    fits_in_memory = True

    def __init__(self,
                 data_directory: str = None,
                 dataset_name: str = None,
                 file_name: str = None,
                 file_directory: str = None,
                 verbose: int = 10,
                 ):
        r"""Initialize a base class of :obj:`MemoryGraphDataset`.

        Args:
            data_directory (str): Full path to directory of the dataset. Default is None.
            file_name (str): Generic filename for dataset to read into memory like a 'csv' file. Default is None.
            file_directory (str): Name or relative path from :obj:`data_directory` to a directory containing sorted
                files. Default is None.
            dataset_name (str): Name of the dataset. Important for naming and saving files. Default is None.
            verbose (int): Logging level. Default is 10.
        """
        super(MemoryGraphDataset, self).__init__()
        # For logging.
        self.logger = logging.getLogger("kgcnn.data." + dataset_name) if dataset_name is not None else module_logger
        self.logger.setLevel(verbose)
        # Dataset information on file.
        self.data_directory = data_directory
        self.file_name = file_name
        self.file_directory = file_directory
        self.dataset_name = dataset_name
        # Data Frame for information.
        self.data_frame = None
        self.data_keys = None
        self.data_unit = None

    @property
    def file_path(self):
        if self.data_directory is None:
            self.warning("Data directory is not set.")
            return None
        if not os.path.exists(self.data_directory):
            self.error("Data directory does not exist.")
        if self.file_name is None:
            self.warning("Can not determine file path.")
            return None
        return os.path.join(self.data_directory, self.file_name)

    @property
    def file_directory_path(self):
        if self.data_directory is None:
            self.warning("Data directory is not set.")
            return None
        if not os.path.exists(self.data_directory):
            self.error("Data directory does not exist.")
        if self.file_directory is None:
            self.warning("Can not determine file directory.")
            return None
        return os.path.join(self.data_directory, self.file_directory)

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.logger.error(*args, **kwargs)

    def save(self, filepath: str = None):
        r"""Save all graph properties to as dictionary as pickled file. By default saves a file named
        :obj:`dataset_name.kgcnn.pickle` in :obj:`data_directory`.

        Args:
            filepath (str): Full path of output file. Default is None.
        """
        if filepath is None:
            filepath = os.path.join(self.data_directory, self.dataset_name + ".kgcnn.pickle")
        self.info("Pickle dataset...")
        save_pickle_file([x.to_dict() for x in self._list], filepath)
        return self

    def load(self, filepath: str = None):
        r"""Load graph properties from a pickled file. By default loads a file named
        :obj:`dataset_name.kgcnn.pickle` in :obj:`data_directory`.

        Args:
            filepath (str): Full path of input file.
        """
        if filepath is None:
            filepath = os.path.join(self.data_directory, self.dataset_name + ".kgcnn.pickle")
        self.info("Load pickled dataset...")
        in_list = load_pickle_file(filepath)
        self._list = [GraphNumpyContainer(x) for x in in_list]
        return self

    def read_in_table_file(self, file_path: str = None, **kwargs):
        r"""Read a data frame in :obj:`data_frame` from file path. By default uses :obj:`file_name` and pandas.
        Checks for a '.csv' file and then for excel file endings. Meaning the file extension of file_path is ignored
        but must be any of the following '.csv', '.xls', '.xlsx', 'odt'.

        Args:
            file_path (str): File path to table file. Default is None.
            kwargs: Kwargs for pandas :obj:`read_csv` function.

        Returns:
            self
        """
        if file_path is None:
            file_path = os.path.join(self.data_directory, self.file_name)
        file_path_base = os.path.splitext(file_path)[0]
        # file_extension_given = os.path.splitext(file_path)[1]

        for file_extension in [".csv"]:
            if os.path.exists(file_path_base + file_extension):
                self.data_frame = pd.read_csv(file_path_base + file_extension, **kwargs)
                return self
        for file_extension in [".xls", ".xlsx", ".xlsm", ".xlsb", ".odf", ".ods", ".odt"]:
            if os.path.exists(file_path_base + file_extension):
                self.data_frame = pd.read_excel(file_path_base + file_extension, **kwargs)
                return self

        self.warning("Unsupported data extension of %s for table file." % file_path)
        return self


MemoryGeometricGraphDataset = MemoryGraphDataset



import numpy as np
import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LinearWarmupExponentialLearningRateScheduler')
class LinearWarmupExponentialLearningRateScheduler(tf.keras.callbacks.LearningRateScheduler):
    """Callback for linear change of the learning rate. This class inherits from
    tf.keras.callbacks.LearningRateScheduler."""

    def __init__(self, lr_start: float, decay_gamma: float, epo_warmup: int = 10, lr_min: float = 0.0,
                 verbose: int = 0):
        """Set the parameters for the learning rate scheduler.

        Args:
            lr_start (float): Learning rate at the start of the exp. decay.
            decay_gamma (float): Gamma parameter in the exponential.
            epo_warmup (int): Number of warm-up steps. Default is 10.
            lr_min (float): Minimum learning rate allowed during the decay. Default is 0.0.
            verbose (int): Verbosity. Default is 0.
        """
        self.decay_gamma = decay_gamma
        self.lr_start = lr_start
        self.lr_min = lr_min
        self.epo_warmup = max(epo_warmup, 0)
        self.verbose = verbose
        super(LinearWarmupExponentialLearningRateScheduler, self).__init__(schedule=self.schedule_epoch_lr,
                                                                           verbose=verbose)

    def schedule_epoch_lr(self, epoch, lr):
        """Reduce the learning rate."""
        if epoch < self.epo_warmup:
            new_lr = self.lr_start * epoch / self.epo_warmup + self.lr_min
        elif epoch == self.epo_warmup:
            new_lr = max(self.lr_start, self.lr_min)
        else:
            new_lr = max(self.lr_start * np.exp(-(epoch - self.epo_warmup) / self.decay_gamma), self.lr_min)
        return float(new_lr)

    def get_config(self):
        config = super(LinearWarmupExponentialLearningRateScheduler, self).get_config()
        config.update({"lr_start": self.lr_start, "decay_gamma": self.decay_gamma, "epo_warmup": self.epo_warmup,
                       "lr_min": self.lr_min, "verbose": self.verbose})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LinearLearningRateScheduler')
class LinearLearningRateScheduler(tf.keras.callbacks.LearningRateScheduler):
    """Callback for linear change of the learning rate. This class inherits from
    tf.keras.callbacks.LearningRateScheduler."""

    def __init__(self, learning_rate_start: float = 1e-3, learning_rate_stop: float = 1e-5, epo_min: int = 0,
                 epo: int = 500, verbose: int = 0):
        """Set the parameters for the learning rate scheduler.

        Args:
            learning_rate_start (float): Initial learning rate. Default is 1e-3.
            learning_rate_stop (float): End learning rate. Default is 1e-5.
            epo_min (int): Minimum number of epochs to keep the learning-rate constant before decrease. Default is 0.
            epo (int): Total number of epochs. Default is 500.
            verbose (int): Verbosity. Default is 0.
        """
        super(LinearLearningRateScheduler, self).__init__(schedule=self.schedule_epoch_lr, verbose=verbose)
        self.learning_rate_start = learning_rate_start
        self.learning_rate_stop = learning_rate_stop
        self.epo = epo
        self.epo_min = epo_min

    def schedule_epoch_lr(self, epoch, lr):
        if epoch < self.epo_min:
            out = float(self.learning_rate_start)
        else:
            out = float(self.learning_rate_start - (self.learning_rate_start - self.learning_rate_stop) / (
                    self.epo - self.epo_min) * (epoch - self.epo_min))
        return float(out)

    def get_config(self):
        config = super(LinearLearningRateScheduler, self).get_config()
        config.update({"learning_rate_start": self.learning_rate_start,
                       "learning_rate_stop": self.learning_rate_stop, "epo": self.epo, "epo_min": self.epo_min})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LinearWarmupExponentialDecay')
class LinearWarmupExponentialDecay(tf.optimizers.schedules.LearningRateSchedule):
    """This schedule combines a linear warmup with an exponential decay."""

    def __init__(self, learning_rate, warmup_steps, decay_steps, decay_rate):
        super().__init__()
        self._input_config_settings = {"learning_rate": learning_rate, "warmup_steps": warmup_steps,
                                       "decay_steps": decay_steps, "decay_rate": decay_rate}
        self.warmup = tf.optimizers.schedules.PolynomialDecay(
            1 / warmup_steps, warmup_steps, end_learning_rate=1)
        self.decay = tf.optimizers.schedules.ExponentialDecay(
            learning_rate, decay_steps, decay_rate)

    def __call__(self, step):
        """Decay lr for step."""
        return self.warmup(step) * self.decay(step)

    def get_config(self):
        """Get config."""
        config = {}
        config.update(self._input_config_settings)
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LearningRateLoggingCallback')
class LearningRateLoggingCallback(tf.keras.callbacks.Callback):
    """Callback logging the learning rate. This does not work universally yet.
    Will be improved in the future."""

    def __init__(self, verbose: int = 0):
        super(LearningRateLoggingCallback, self).__init__()
        self.verbose = verbose

    def on_epoch_end(self, epoch, logs=None):
        """Read out the learning rate on epoch end.

        Args:
            epoch (int): Number of current epoch ended.
            logs (dict): Dictionary of the logs.

        Returns:
            None.
        """
        lr = self.model.optimizer.lr
        tf.summary.scalar('learning rate', data=lr, step=epoch)
        logs = logs or {}
        logs['lr'] = tf.keras.backend.get_value(self.model.optimizer.lr)
        if self.verbose > 0:
            print('\nEpoch %05d: LearningRateScheduler reducing learning '
                  'rate to %s.' % (epoch + 1, lr))

    def get_config(self):
        """Get config."""
        config = {"verbose": self.verbose}
        return config



import numpy as np
import tensorflow as tf

from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.gather import GatherNodesSelection, GatherState
from kgcnn.layers.modules import LazySubtract, LazyMultiply
from kgcnn.ops.axis import get_positive_axis


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='NodePosition')
class NodePosition(GraphBaseLayer):
    """Get node position for edges. Directly calls `GatherNodesSelection` with provided index tensor.
    Returns separate node position tensor for each of the indices. Index selection must be provided
    in the constructor. Defaults to first two indices of an edge.

    A distance based edge is defined by two bond indices of the index list of shape (batch, [M], 2) with last dimension
    of incoming and outgoing node (message passing framework).
    """

    def __init__(self, selection_index: list = None, **kwargs):
        """Initialize layer instance of `NodePosition`.

        Args:
            selection_index (list): List of positions (last dimension of the index tensor) to return node coordinates.
                Default is [0, 1].
        """
        super(NodePosition, self).__init__(**kwargs)
        if selection_index is None:
            selection_index = [0, 1]
        self.selection_index = selection_index
        self.layer_gather = GatherNodesSelection(self.selection_index)

    def build(self, input_shape):
        """Build layer."""
        super(NodePosition, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass of `NodePosition`.

        Args:
            inputs (list): [position, edge_index]

                - position (tf.RaggedTensor): Node positions of shape (batch, [N], 3).
                - edge_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2).

        Returns:
            list: List of node positions (ragged) tensors for each of the `selection_index`. Position tensors have
                shape (batch, [M], 3).
        """
        return self.layer_gather(inputs, **kwargs)

    def get_config(self):
        """Update config for `NodePosition`."""
        config = super(NodePosition, self).get_config()
        config.update({"selection_index": self.selection_index})
        return config


class ShiftPeriodicLattice(GraphBaseLayer):
    """Node position periodic.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(ShiftPeriodicLattice, self).__init__(**kwargs)
        self.layer_state = GatherState()

    def build(self, input_shape):
        """Build layer."""
        super(ShiftPeriodicLattice, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [position, edge_index]

                - position (tf.RaggedTensor): Positions of shape (batch, [M], 3)
                - edge_image (tf.RaggedTensor): Position in which image to shift of shape (batch, [M], 3)
                - lattice (tf.tensor): Lattice vector matrix of shape (batch, 3, 3)

        Returns:
            tf.RaggedTensor: Gathered node position number of indices of shape (batch, [M], 1)
        """
        self.assert_ragged_input_rank(inputs[:2])
        lattice_rep = self.layer_state([inputs[2], inputs[1]], **kwargs)  # Should be (batch, None, 3, 3)
        x = inputs[0]
        xj = x.values
        xj = xj + tf.reduce_sum(tf.cast(lattice_rep.values, dtype=xj.dtype) * tf.expand_dims(
            tf.cast(inputs[1].values, dtype=xj.dtype), axis=-1), axis=1)
        return tf.RaggedTensor.from_row_splits(xj, inputs[1].row_splits, validate=self.ragged_validate)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EuclideanNorm')
class EuclideanNorm(GraphBaseLayer):
    """Compute euclidean norm for edge or node vectors. This amounts for a specific axis to

    :math:`||x||_2 = \sqrt{\sum_i x_i^2}`

    Vector based edge or node coordinates are defined by (batch, [N], ..., D) with last dimension D.
    You can choose to collapse or keep the dimension.
    """

    def __init__(self, axis: int = -1, keepdims: bool = False, invert_norm: bool = False, **kwargs):
        """Initialize layer."""
        super(EuclideanNorm, self).__init__(**kwargs)
        self.axis = axis
        self.keepdims = keepdims
        self.invert_norm = invert_norm

    def build(self, input_shape):
        """Build layer."""
        super(EuclideanNorm, self).build(input_shape)

    @staticmethod
    def _compute_euclidean_norm(inputs, axis: int = -1, keepdims: bool = False, invert_norm: bool = False):
        """Function to compute euclidean norm for inputs."""
        out = tf.sqrt(tf.nn.relu(tf.reduce_sum(tf.square(inputs), axis=axis, keepdims=keepdims)))
        if invert_norm:
            out = tf.math.divide_no_nan(tf.constant(1, dtype=out.dtype), out)
        return out

    def call(self, inputs, **kwargs):
        """Forward pass for `EuclideanNorm`.

        Args:
            inputs (tf.RaggedTensor): Positions of shape (batch, [N], ..., D, ...)

        Returns:
            tf.RaggedTensor: Euclidean norm computed for specific axis of shape (batch, [N], ...)
        """
        # Possibly faster
        if isinstance(inputs, tf.RaggedTensor):
            axis = get_positive_axis(self.axis, inputs.shape.rank)
            if inputs.ragged_rank == 1 and axis > 1:
                out = self._compute_euclidean_norm(inputs.values, axis-1, self.keepdims, self.invert_norm)
                return tf.RaggedTensor.from_row_splits(out, inputs.row_splits, validate=self.ragged_validate)
        return self._compute_euclidean_norm(inputs, self.axis, self.keepdims, self.invert_norm)

    def get_config(self):
        """Update config."""
        config = super(EuclideanNorm, self).get_config()
        config.update({"axis": self.axis, "keepdims": self.keepdims, "invert_norm": self.invert_norm})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ScalarProduct')
class ScalarProduct(GraphBaseLayer):
    """Compute geometric scalar product for edge or node coordinates.

    A distance based edge or node coordinates are defined by (batch, [N], ..., D) with last dimension D.
    """

    def __init__(self, axis=-1, **kwargs):
        """Initialize layer."""
        super(ScalarProduct, self).__init__(**kwargs)
        self.axis = axis

    def build(self, input_shape):
        """Build layer."""
        super(ScalarProduct, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [vec1, vec2]

                - vec1 (tf.RaggedTensor): Positions of shape (batch, [N], ..., D, ...)
                - vec2 (tf.RaggedTensor): Positions of shape (batch, [N], ..., D, ...)

        Returns:
            tf.RaggedTensor: Scalar product of shape (batch, [N], ...)
        """
        if all([isinstance(x, tf.RaggedTensor) for x in inputs]):  # Possibly faster
            axis = get_positive_axis(self.axis, inputs[0].shape.rank)
            axis2 = get_positive_axis(self.axis, inputs[1].shape.rank)
            assert axis2 == axis
            if all([x.ragged_rank == 1 for x in inputs]) and axis > 1:
                v1, v2 = inputs[0].values, inputs[1].values
                out = tf.reduce_sum(v1 * v2, axis=axis - 1)
                return tf.RaggedTensor.from_row_splits(out, inputs[0].row_splits, validate=self.ragged_validate)
        # Default
        out = tf.reduce_sum(inputs[0] * inputs[1], axis=self.axis)
        return out

    def get_config(self):
        """Update config."""
        config = super(ScalarProduct, self).get_config()
        config.update({"axis": self.axis})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='NodeDistanceEuclidean')
class NodeDistanceEuclidean(GraphBaseLayer):
    r"""Compute euclidean distance between two node coordinate tensors. Let the :math:`\vec{x}_1` and :math:`\vec{x}_2`
    be two nodes, then the output is given by :math:`|| \vec{x}_1 - \vec{x}_2 ||_2`. Calls :obj:`EuclideanNorm` on
    the difference of the inputs. The number of node positions must not be connected to the number of nodes, but can
    also match edges.
    """

    def __init__(self, **kwargs):
        """Initialize layer instance of `NodeDistanceEuclidean`."""
        super(NodeDistanceEuclidean, self).__init__(**kwargs)
        self.layer_subtract = LazySubtract()
        self.layer_euclidean_norm = EuclideanNorm(axis=2, keepdims=True)

    def build(self, input_shape):
        """Build layer."""
        super(NodeDistanceEuclidean, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [position_start, position_stop]

                - position_start (tf.RaggedTensor): Node positions of shape (batch, [M], 3)
                - position_stop (tf.RaggedTensor): Node positions of shape (batch, [M], 3)

        Returns:
            tf.RaggedTensor: Distances as edges that match the number of indices of shape (batch, [M], 1)
        """
        diff = self.layer_subtract(inputs)
        return self.layer_euclidean_norm(diff)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EdgeDirectionNormalized')
class EdgeDirectionNormalized(GraphBaseLayer):
    r"""Compute the normalized geometric direction between two point coordinates for e.g. a geometric edge.

    :math:`\frac{\vec{r}_{ij}}{||{r_ij}||} = \frac{\vec{r}_{i} - \vec{r}_{j}}{||\vec{r}_{i} - \vec{r}_{j}||}`.

    Note that the difference is defined here as :math:`\vec{r}_{i} - \vec{r}_{j}`.
    As the first index defines the incoming edge.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(EdgeDirectionNormalized, self).__init__(**kwargs)
        self.layer_subtract = LazySubtract()
        self.layer_euclidean_norm = EuclideanNorm(axis=2, keepdims=True, invert_norm=True)
        self.layer_multiply = LazyMultiply()

    def build(self, input_shape):
        super(EdgeDirectionNormalized, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [position_1, position_2]

                - position_1 (tf.RaggedTensor): Stop node positions of shape (batch, [N], 3)
                - position_2 (tf.RaggedTensor): Start node positions of shape (batch, [N], 3)

        Returns:
            tf.RaggedTensor: Normalized vector distance of shape (batch, [M], 3).
        """
        diff = self.layer_subtract(inputs)
        norm = self.layer_euclidean_norm(diff)
        return self.layer_multiply([diff, norm])

    def get_config(self):
        """Update config."""
        config = super(EdgeDirectionNormalized, self).get_config()
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='VectorAngle')
class VectorAngle(GraphBaseLayer):
    r"""Compute geometric angles between vectors in euclidean space. The vectors would be obtained from the points
    :math:`x_i` from :math:`v_1 = x_i - x_j` and :math:`v_2 = x_j - x_k`.

    The geometric angle is computed between i<-j,j<-k for index tuple (i,j,k) in (batch, None, 3) last dimension.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(VectorAngle, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(VectorAngle, self).build(input_shape)

    @staticmethod
    def _compute_vector_angle(inputs):
        """Function to compute angles between v1 and v2"""
        v1, v2 = inputs[0], inputs[1]
        x = tf.reduce_sum(v1 * v2, axis=-1)
        y = tf.linalg.cross(v1, v2)
        y = tf.norm(y, axis=-1)
        angle = tf.math.atan2(y, x)
        out = tf.expand_dims(angle, axis=-1)
        return out

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [vector_1, vector_2]

                - vector_1 (tf.RaggedTensor): Node positions or vectors of shape (batch, [M], 3)
                - vector_2 (tf.RaggedTensor): Node positions or vectors of shape (batch, [M], 3)

        Returns:
            tf.RaggedTensor: Calculate Angle between vector 1 and 2 of shape (batch, [M], 1)
        """
        if all([isinstance(x, tf.RaggedTensor) for x in inputs]):  # Possibly faster
            if all([x.ragged_rank == 1 for x in inputs]):
                angle = self._compute_vector_angle([inputs[0].values, inputs[1].values])
                return tf.RaggedTensor.from_row_splits(angle, inputs[0].row_splits, validate=self.ragged_validate)
        return self._compute_vector_angle(inputs)

    def get_config(self):
        """Update config."""
        config = super(VectorAngle, self).get_config()
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EdgeAngle')
class EdgeAngle(GraphBaseLayer):
    """Compute geometric edge angles.

    The geometric angle is computed between edge tuple of index (i,j), where i,j refer to two edges.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(EdgeAngle, self).__init__(**kwargs)
        self.layer_gather_vectors = GatherNodesSelection([0, 1])
        self.layer_angle = VectorAngle()

    def build(self, input_shape):
        """Build layer."""
        super(EdgeAngle, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [vector, angle_index]

                - vector (tf.RaggedTensor): Node or Edge directions of shape (batch, [N], 3)
                - angle_index (tf.RaggedTensor): Angle indices of vector pairs of shape (batch, [K], 2).

        Returns:
            tf.RaggedTensor: Edge angles between edges that match the indices. Shape is (batch, [K], 1)
        """
        v1, v2 = self.layer_gather_vectors(inputs)
        return self.layer_angle([v1, v2])

    def get_config(self):
        """Update config."""
        config = super(EdgeAngle, self).get_config()
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GaussBasisLayer')
class GaussBasisLayer(GraphBaseLayer):
    r"""Expand a distance into a Gauss Basis with :math:`\sgima`, according to Schuett et al."""

    def __init__(self, bins=20, distance=4.0, sigma=0.4, offset=0.0,
                 **kwargs):
        super(GaussBasisLayer, self).__init__(**kwargs)
        # Layer variables
        self.bins = int(bins)
        self.distance = float(distance)
        self.offset = float(offset)
        self.sigma = float(sigma)
        self.gamma = 1 / sigma / sigma * (-1) / 2
        # Note: For arbitrary axis the code must be adapted.

    @staticmethod
    def _compute_gauss_basis(inputs, offset, gamma, bins, distance):
        """expand into gaussian basis."""
        gbs = tf.range(0, bins, 1, dtype=inputs.dtype) / float(bins) * distance
        out = inputs - offset
        out = tf.square(out - gbs) * gamma
        out = tf.exp(out)
        return out

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: distance

                - distance (tf.RaggedTensor): Edge distance of shape (batch, [K], 1)

        Returns:
            tf.RaggedTensor: Expanded distance. Shape is (batch, [K], #bins)
        """
        # Possibly faster RaggedRank==1
        if isinstance(inputs, tf.RaggedTensor):
            if inputs.ragged_rank == 1:
                out = self._compute_gauss_basis(inputs.values, self.offset, self.gamma, self.bins, self.distance)
                return tf.RaggedTensor.from_row_splits(out, inputs.row_splits, validate=self.ragged_validate)
        return self._compute_gauss_basis(inputs, self.offset, self.gamma, self.bins, self.distance)

    def get_config(self):
        """Update config."""
        config = super(GaussBasisLayer, self).get_config()
        config.update({"bins": self.bins, "distance": self.distance, "offset": self.offset, "sigma": self.sigma})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='BesselBasisLayer')
class BesselBasisLayer(GraphBaseLayer):
    r"""Expand a distance into a Bessel Basis with :math:`l=m=0`, according to Klicpera et al. 2020

    Args:
        num_radial (int): Number of radial radial basis functions
        cutoff (float): Cutoff distance c
        envelope_exponent (int): Degree of the envelope to smoothen at cutoff. Default is 5.
    """

    def __init__(self, num_radial,
                 cutoff,
                 envelope_exponent=5,
                 envelope_type="poly",
                 **kwargs):
        super(BesselBasisLayer, self).__init__(**kwargs)
        # Layer variables
        self.num_radial = num_radial
        self.cutoff = cutoff
        self.inv_cutoff = tf.constant(1 / cutoff, dtype=tf.float32)
        self.envelope_exponent = envelope_exponent
        self.envelope_type = str(envelope_type)

        if self.envelope_type not in ["poly"]:
            raise ValueError("Unknown envelope type %s in BesselBasisLayer" % self.envelope_type)

        # Initialize frequencies at canonical positions
        def freq_init(shape, dtype):
            return tf.constant(np.pi * np.arange(1, shape + 1, dtype=np.float32), dtype=dtype)

        self.frequencies = self.add_weight(name="frequencies", shape=self.num_radial,
                                           dtype=tf.float32, initializer=freq_init, trainable=True)

    @tf.function
    def envelope(self, inputs):
        p = self.envelope_exponent + 1
        a = -(p + 1) * (p + 2) / 2
        b = p * (p + 2)
        c = -p * (p + 1) / 2
        env_val = 1.0 / inputs + a * inputs ** (p - 1) + b * inputs ** p + c * inputs ** (p + 1)
        return tf.where(inputs < 1, env_val, tf.zeros_like(inputs))

    def expand_bessel_basis(self, inputs):
        d_scaled = inputs * self.inv_cutoff
        d_cutoff = self.envelope(d_scaled)
        out = d_cutoff * tf.sin(self.frequencies * d_scaled)
        return out

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: distance

                - distance (tf.RaggedTensor): Edge distance of shape (batch, [K], 1)

        Returns:
            tf.RaggedTensor: Expanded distance. Shape is (batch, [K], #Radial)
        """
        # Possibly faster RaggedRank==1
        if isinstance(inputs, tf.RaggedTensor):
            if inputs.ragged_rank == 1:
                out = self.expand_bessel_basis(inputs.values)
                return tf.RaggedTensor.from_row_splits(out, inputs.row_splits, validate=self.ragged_validate)
        return self.expand_bessel_basis(inputs)

    def get_config(self):
        """Update config."""
        config = super(BesselBasisLayer, self).get_config()
        config.update({"num_radial": self.num_radial, "cutoff": self.cutoff,
                       "envelope_exponent": self.envelope_exponent, "envelope_type": self.envelope_type})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='CosCutOffEnvelope')
class CosCutOffEnvelope(GraphBaseLayer):
    r"""Calculate cos-cutoff envelope according to Behler et al. https://aip.scitation.org/doi/10.1063/1.3553717
    :math:`f_c(R_{ij}) = 0.5 [ \cos{\frac{\pi R_{ij}}{R_c}} + 1]`

    Args:
        cutoff (float): Cutoff distance :math:`R_c`.
    """

    def __init__(self,
                 cutoff,
                 **kwargs):
        super(CosCutOffEnvelope, self).__init__(**kwargs)
        self.cutoff = float(np.abs(cutoff)) if cutoff is not None else 1e8

    @staticmethod
    def _compute_cutoff_envelope(inputs, cutoff):
        """Implements the cutoff envelope."""
        fc = tf.clip_by_value(inputs, -cutoff, cutoff)
        fc = (tf.math.cos(fc * np.pi / cutoff) + 1) * 0.5
        # fc = tf.where(tf.abs(inputs) < self.cutoff, fc, tf.zeros_like(fc))
        return fc

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: distance

                - distance (tf.RaggedTensor): Edge distance of shape (batch, [M], 1)

        Returns:
            tf.RaggedTensor: Cutoff envelope of shape (batch, [M], 1)
        """
        return self.call_on_values_tensor_of_ragged(self._compute_cutoff_envelope, inputs, cutoff=self.cutoff)

    def get_config(self):
        """Update config."""
        config = super(CosCutOffEnvelope, self).get_config()
        config.update({"cutoff": self.cutoff})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='CosCutOff')
class CosCutOff(GraphBaseLayer):
    r"""Apply cos-cutoff according to Behler et al. https://aip.scitation.org/doi/10.1063/1.3553717
    :math:`f_c(R_{ij}) = 0.5 [ \cos{\frac{\pi R_{ij}}{R_c}} + 1]` by simply multiplying with the envelope.

    Args:
        cutoff (float): Cutoff distance :math:`R_c`.
    """

    def __init__(self,
                 cutoff,
                 **kwargs):
        super(CosCutOff, self).__init__(**kwargs)
        self.cutoff = float(np.abs(cutoff)) if cutoff is not None else 1e8

    @staticmethod
    def _compute_cutoff(inputs, cutoff):
        fc = tf.clip_by_value(inputs, -cutoff, cutoff)
        fc = (tf.math.cos(fc * np.pi / cutoff) + 1) * 0.5
        # fc = tf.where(tf.abs(inputs) < self.cutoff, fc, tf.zeros_like(fc))
        out = fc * inputs
        return out

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: distance
                - distance (tf.RaggedTensor): Edge distance of shape (batch, [M], D)

        Returns:
            tf.RaggedTensor: Cutoff applied to input of shape (batch, [M], D)
        """
        return self.call_on_values_tensor_of_ragged(self._compute_cutoff, inputs, cutoff=self.cutoff)

    def get_config(self):
        """Update config."""
        config = super(CosCutOff, self).get_config()
        config.update({"cutoff": self.cutoff})
        return config



import importlib


class ModelSelection:
    """Select a model by string identifier. Since most models are created by the functional API, we do not register
    them but have an import by identifier here."""

    def __init__(self, model_name: str = None):
        """Initialize a Model-selection for a specific model.

        Args:
            model_name (str):  Model name. Default is None
        """
        self._model_name = None
        if model_name is not None:
            self._model_name = model_name

    def make_model(self, model_id: str = None):
        """Return a make_model function that generates model instances from model-specific hyper-parameters.

        Args:
            model_id (str): Name of the model.

        Returns:
            func: Function to make model instances of tf.keras.Model.
        """
        if model_id is None:
            model_id = self._model_name

        try:
            make_model = getattr(importlib.import_module("kgcnn.literature.%s" % model_id), "make_model")

        except ModuleNotFoundError:
            raise NotImplementedError("Unknown model identifier %s for a model in kgcnn.literature" % model_id)

        return make_model

    def __call__(self, **model_kwargs):
        """Make a model for with model kwargs.

        Args:
            model_kwargs (dict): Model kwargs.

        Returns:
            tf.keras.models.Model
        """
        return self.make_model()(**model_kwargs)


import os
import numpy as np
import pandas as pd
import logging

from kgcnn.mol.gen.base import smile_to_mol
from kgcnn.data.base import MemoryGraphDataset
from kgcnn.utils.data import save_json_file, load_json_file
from kgcnn.mol.graphRD import MolecularGraphRDKit
from kgcnn.mol.enocder import OneHotEncoder
from kgcnn.mol.io import write_mol_block_list_to_sdf, read_mol_list_from_sdf_file
from kgcnn.utils.data import pandas_data_frame_columns_to_numpy


class MoleculeNetDataset(MemoryGraphDataset):
    r"""Class for using molecule datasets. The concept is to load a table of smiles and corresponding targets and
    convert them into a tensors representation for graph networks.

    The class provides properties and methods for making graph features from smiles.
    The typical input is a `csv` or `excel` file with smiles and corresponding graph labels.

    The graph structure matches the molecular graph, i.e. the chemical structure. The atomic coordinates
    are generated by a conformer guess. Since this require some computation time, it is only done once and the
    molecular coordinate or mol-blocks stored in a single SDF file with the base-name of the csv :obj:``file_name``.

    The selection of smiles and whether conformers should be generated is handled by sub-classes or specified in the
    the methods :obj:`prepare_data` and :obj:`read_in_memory`, see the documentation of the methods
    for further details.
    """

    def __init__(self, data_directory: str = None, dataset_name: str = None, file_name: str = None,
                 verbose: int = 10):
        r"""Initialize a `MoleculeNetDataset` with information of the dataset location on disk.

        Args:
            file_name (str): Filename for reading into memory. This must be the name of the '.csv' file.
                Default is None.
            data_directory (str): Full path to directory containing all dataset files. Default is None.
            dataset_name (str): Name of the dataset. Important for naming. Default is None.
            verbose (int): Logging level. Default is 10.
        """
        MemoryGraphDataset.__init__(self, data_directory=data_directory, dataset_name=dataset_name,
                                    file_name=file_name, verbose=verbose)

    @property
    def file_path_mol(self):
        """Try to determine a file path for the mol information to store."""
        return os.path.splitext(self.file_path)[0] + ".sdf"

    def _smiles_to_mol_list(self, smiles: list, add_hydrogen: bool = True, sanitize: bool = True,
                            make_conformers: bool = True, optimize_conformer: bool = True,
                            external_program: dict = None, num_workers: int = None):
        r"""Convert a list of smiles as string into a list of mol-information, namely mol-block as string.

        Args:
            smiles (list): A list of smiles for each molecule in dataset.
            add_hydrogen (bool): Whether to add hydrogen after smile translation.
            sanitize (bool): Whether to sanitize molecule.
            make_conformers (bool): Try to generate 3D coordinates
            optimize_conformer (bool): Whether to optimize conformer via force field.
                Only possible with :obj:`make_conformers`. Default is True.
            external_program (dict): External program for translating smiles. Default is None.
                If you want to use an external program you have to supply a dictionary of the form:
                {"class_name": "balloon", "config": {"balloon_executable_path": ..., ...}}.
                Note that usually the parameters like :obj:`add_hydrogen` are ignored. And you need to control the
                SDF file generation within `config` of the :obj:`external_program`.
            num_workers (int): Parallel execution for translating smiles.

        Returns:
            list: A list of mol-block information as sting.
        """
        if len(smiles) == 0:
            self.error("Can not translate smiles, received empty list for %s." % self.dataset_name)

        self.info("Generating molecules and store %s to disk..." % self.file_path_mol)
        molecule_list = []
        for i in range(0, len(smiles), 1000):
            mg = smile_to_mol(smiles[i:i+1000], self.data_directory,
                              add_hydrogen=add_hydrogen, sanitize=sanitize,
                              make_conformers=make_conformers, optimize_conformer=optimize_conformer,
                              external_program=external_program, num_workers=num_workers)
            molecule_list = molecule_list + mg
            self.info(" ... converted molecules {0} from {1}".format(i+len(mg), len(smiles)))

        return molecule_list

    def prepare_data(self, overwrite: bool = False, smiles_column_name: str = "smiles",
                     add_hydrogen: bool = True, sanitize: bool = True,
                     make_conformers: bool = True, optimize_conformer: bool = True,
                     external_program: dict = None, num_workers: int = None):
        r"""Pre-computation of molecular structure information and optionally conformers. This function reads smiles
        from the csv-file given by :obj:`file_name` and creates a SDF File of generated mol-blocks with the same
        file name. The class requires RDKit and OpenBabel.
        Smiles that are not compatible with both RDKit and OpenBabel result in an empty mol-block in the SDF file to
        keep the number of molecules the same.

        Args:
            overwrite (bool): Overwrite existing database mol-json file. Default is False.
            smiles_column_name (str): Column name where smiles are given in csv-file. Default is "smiles".
            add_hydrogen (bool): Whether to add H after smile translation. Default is True.
            sanitize (bool): Whether to sanitize molecule.
            make_conformers (bool): Whether to make conformers. Default is True.
            optimize_conformer (bool): Whether to optimize conformer via force field.
                Only possible with :obj:`make_conformers`. Default is True.
            external_program (dict): External program for translating smiles. Default is None.
                If you want to use an external program you have to supply a dictionary of the form:
                {"class_name": "balloon", "config": {"balloon_executable_path": ..., ...}}.
                Note that usually the parameters like :obj:`add_hydrogen` are ignored. And you need to control the
                SDF file generation within `config` of the :obj:`external_program`.
            num_workers (int): Parallel execution for translating smiles.

        Returns:
            self
        """
        if os.path.exists(self.file_path_mol) and not overwrite:
            self.info("Found SDF %s of pre-computed structures." % self.file_path_mol)
            return self

        self.read_in_table_file()
        smiles = self.data_frame[smiles_column_name].values

        mb = self._smiles_to_mol_list(smiles,
                                      add_hydrogen=add_hydrogen, sanitize=sanitize,
                                      make_conformers=make_conformers, optimize_conformer=optimize_conformer,
                                      external_program=external_program, num_workers=num_workers)

        write_mol_block_list_to_sdf(mb, self.file_path_mol)
        return self

    def read_in_memory(self,  label_column_name: str = None, has_conformers: bool = True,
                       add_hydrogen: bool = True):
        r"""Load list of molecules from cached SDF-file in into memory. File name must be given in :obj:`file_name` and
        path information in the constructor of this class. Extract basic graph information from mol-blocks.
        No further attributes are computed as default. Use :obj:`set_attributes` for this purpose.
        It further checks the csv-file for graph labels specified by :obj:`label_column_name`.
        Labels that do not have valid smiles or molecule in the SDF-file are also skipped.
        The assignment to original IDs is stored in :obj:`valid_molecule_id`.

        Args:
            label_column_name (str): Column name in the csv-file where to take graph labels from.
            add_hydrogen (bool): Whether to keep hydrogen after reading the mol-information. Default is True.
            has_conformers (bool): If molecules has 3D coordinates pre-computed.
                For multi-targets you can supply a list of column names or positions. Also a slice can be provided
                for selecting columns as graph labels. Default is None.

        Returns:
            self
        """
        # Read the data from a csv-file.
        data = pd.read_csv(os.path.join(self.data_directory, self.file_name))

        # Find columns to take graph labels from.
        self.data_keys = data.columns
        graph_labels_all = pandas_data_frame_columns_to_numpy(data, label_column_name)

        if not os.path.exists(self.file_path_mol):
            raise FileNotFoundError("Can not load molecules for dataset %s" % self.dataset_name)

        self.info("Read mol-blocks from %s of pre-computed structures..." % self.file_path_mol)
        mols = read_mol_list_from_sdf_file(self.file_path_mol)

        # Main loop to read molecules from mol-block
        node_symbol = []
        node_coordinates = []
        node_number = []
        edge_indices = []
        edge_number = []
        num_mols = len(mols)
        graph_labels = []
        counter_iter = 0
        for i, x in enumerate(mols):
            mg = MolecularGraphRDKit(add_hydrogen=add_hydrogen).from_mol_block(x, sanitize=True)
            if mg.mol is None:
                edge_indices.append(None)
                edge_number.append(None)
                node_symbol.append(None)
                node_number.append(None)
                graph_labels.append(None)
                node_coordinates.append(None)
                continue
            # Append all valid tensor quantities
            temp_edge = mg.edge_number
            edge_indices.append(temp_edge[0])
            edge_number.append(np.array(temp_edge[1], dtype="int"))
            node_symbol.append(mg.node_symbol)
            node_number.append(mg.node_number)
            graph_labels.append(graph_labels_all[i])
            node_coordinates.append(mg.node_coordinates)
            counter_iter += 1
            if i % 1000 == 0:
                self.info(" ... read molecules {0} from {1}".format(i, num_mols))

        self.assign_property("node_symbol", node_symbol)
        self.assign_property("node_coordinates",  node_coordinates)
        self.assign_property("node_number", node_number)
        self.assign_property("graph_size", [None if x is None else len(x) for x in node_number])
        self.assign_property("edge_indices", edge_indices)
        self.assign_property("graph_labels", graph_labels)
        self.assign_property("edge_number", edge_number)
        return self

    def set_attributes(self,
                       nodes: list = None,
                       edges: list = None,
                       graph: list = None,
                       encoder_nodes: dict = None,
                       encoder_edges: dict = None,
                       encoder_graph: dict = None,
                       add_hydrogen: bool = False,
                       has_conformers: bool = True,
                       ):
        r"""Set further molecular attributes or features by string identifier. Requires :obj:`MolecularGraphRDKit`.
        Reset edges and nodes with new attributes and edge indices. Default values are features that has been used
        by `Luo et al (2019) <https://doi.org/10.1021/acs.jmedchem.9b00959>`_.

        Args:
            nodes (list): A list of node attributes
            edges (list): A list of edge attributes
            graph (list): A list of graph attributes.
            encoder_nodes (dict): A dictionary of callable encoder where the key matches the attribute.
            encoder_edges (dict): A dictionary of callable encoder where the key matches the attribute.
            encoder_graph (dict): A dictionary of callable encoder where the key matches the attribute.
            add_hydrogen (bool): Whether to remove hydrogen.
            has_conformers (bool): If molecules needs 3D coordinates pre-computed.

        Returns:
            self
        """
        if not os.path.exists(self.file_path_mol):
            raise FileNotFoundError("Can not load molecules for dataset %s" % self.dataset_name)

        self.info("Making attributes...")

        mols = read_mol_list_from_sdf_file(self.file_path_mol)

        # Choose default values here:
        if nodes is None:
            nodes = ['Symbol', 'TotalDegree', 'FormalCharge', 'NumRadicalElectrons', 'Hybridization',
                     'IsAromatic', 'IsInRing', 'TotalNumHs', 'CIPCode', "ChiralityPossible", "ChiralTag"]
        if edges is None:
            edges = ['BondType', 'IsAromatic', 'IsConjugated', 'IsInRing', "Stereo"]
        if graph is None:
            graph = ['ExactMolWt', 'NumAtoms']
        if encoder_nodes is None:
            encoder_nodes = {
                "Symbol": OneHotEncoder(
                    ['B', 'C', 'N', 'O', 'F', 'Si', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Te', 'I', 'At'], dtype="str"),
                "Hybridization": OneHotEncoder([2, 3, 4, 5, 6]),
                "TotalDegree": OneHotEncoder([0, 1, 2, 3, 4, 5], add_unknown=False),
                "TotalNumHs": OneHotEncoder([0, 1, 2, 3, 4], add_unknown=False),
                "CIPCode": OneHotEncoder(['R', 'S'], add_unknown=False, dtype="str")
            }
        if encoder_edges is None:
            encoder_edges = {
                "BondType": OneHotEncoder([1, 2, 3, 12], add_unknown=False),
                "Stereo": OneHotEncoder([0, 1, 2, 3], add_unknown=False),
            }
        if encoder_graph is None:
            encoder_graph = {}

        for key, value in encoder_nodes.items():
            encoder_nodes[key] = self._deserialize_encoder(value)
        for key, value in encoder_edges.items():
            encoder_edges[key] = self._deserialize_encoder(value)
        for key, value in encoder_graph.items():
            encoder_graph[key] = self._deserialize_encoder(value)

        # Reset all attributes
        graph_attributes = []
        node_attributes = []
        edge_attributes = []
        edge_number = []
        edge_indices = []
        node_coordinates = []
        node_symbol = []
        node_number = []
        num_mols = len(mols)
        counter_iter = 0
        for i, sm in enumerate(mols):
            mg = MolecularGraphRDKit(add_hydrogen=add_hydrogen).from_mol_block(sm, sanitize=True)
            if mg.mol is None:
                node_coordinates.append(None)
                edge_indices.append(None)
                edge_number.append(None)
                node_attributes.append(None)
                edge_attributes.append(None)
                graph_attributes.append(None)
                node_symbol.append(None)
                node_number.append(None)
                continue
            # Append all valid tensor properties
            temp_edge = mg.edge_number
            node_coordinates.append(np.array(mg.node_coordinates, dtype="float32"))
            edge_indices.append(np.array(temp_edge[0], dtype="int64"))
            edge_number.append(np.array(temp_edge[1], dtype="int"))
            node_attributes.append(np.array(mg.node_attributes(nodes, encoder_nodes), dtype="float32"))
            edge_attributes.append(np.array(mg.edge_attributes(edges, encoder_edges)[1], dtype="float32"))
            graph_attributes.append(np.array(mg.graph_attributes(graph, encoder_graph), dtype="float32"))
            node_symbol.append(mg.node_symbol)
            node_number.append(mg.node_number)
            counter_iter += 1
            if i % 1000 == 0:
                self.info(" ... read molecules {0} from {1}".format(i, num_mols))

        self.assign_property("graph_size", [None if x is None else len(x) for x in node_number])
        self.assign_property("graph_attributes", graph_attributes)
        self.assign_property("node_attributes", node_attributes)
        self.assign_property("edge_attributes", edge_attributes)
        self.assign_property("edge_indices", edge_indices)
        self.assign_property("node_coordinates", node_coordinates)
        self.assign_property("node_symbol", node_symbol)
        self.assign_property("node_number", node_number)

        if self.logger.getEffectiveLevel() < 20:
            for key, value in encoder_nodes.items():
                if hasattr(value, "report"):
                    value.report(name=key)
            for key, value in encoder_edges.items():
                if hasattr(value, "report"):
                    value.report(name=key)
            for key, value in encoder_graph.items():
                if hasattr(value, "report"):
                    value.report(name=key)

        return self

    @staticmethod
    def _deserialize_encoder(encoder_identifier):
        """Serialization. Will maybe include keras in the future.

        Args:
            encoder_identifier: Identifier, class or function of an encoder.

        Returns:
            obj: Deserialized encoder.
        """
        if isinstance(encoder_identifier, dict):
            if encoder_identifier["class_name"] == "OneHotEncoder":
                return OneHotEncoder.from_config(encoder_identifier["config"])
        elif hasattr(encoder_identifier, "__call__"):
            return encoder_identifier
        else:
            raise ValueError("Unable to deserialize encoder %s " % encoder_identifier)



import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.layers.base import GraphBaseLayer
from kgcnn.ops.axis import get_positive_axis


# There are limitations for RaggedTensor working with standard Keras layers. Here are some simple surrogates.
# This is a temporary solution until future versions of TensorFlow support more RaggedTensor arguments.
# Since most kgcnn layers work with ragged_rank = 1 and defined inner dimension, this case can be caught explicitly.


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='DenseEmbedding')
class DenseEmbedding(GraphBaseLayer):
    r"""Dense layer for ragged tensors representing a geometric or graph tensor such as node or edge embeddings.
    Current tensorflow version now support ragged input for :obj:`tf.keras.layers.Dense` with defined inner dimension.
    This layer is kept for backward compatibility and but does not necessarily have to be used in models anymore.
    """

    def __init__(self,
                 units,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        """Initialize layer as tf.keras.Dense."""
        super(DenseEmbedding, self).__init__(**kwargs)
        self._layer_dense = ks.layers.Dense(units=units, activation=activation,
                                            use_bias=use_bias, kernel_initializer=kernel_initializer,
                                            bias_initializer=bias_initializer,
                                            kernel_regularizer=kernel_regularizer,
                                            bias_regularizer=bias_regularizer,
                                            activity_regularizer=activity_regularizer,
                                            kernel_constraint=kernel_constraint,
                                            bias_constraint=bias_constraint)
        self._add_layer_config_to_self = {
            "_layer_dense": ["units", "activation", "use_bias", "kernel_initializer", "bias_initializer",
                             "kernel_regularizer", "bias_regularizer", "activity_regularizer",
                             "kernel_constraint", "bias_constraint"]}

    def call(self, inputs, **kwargs):
        """Forward pass using Dense on flat values."""
        # For Dense can call on flat values.
        if isinstance(inputs, tf.RaggedTensor):
            return tf.ragged.map_flat_values(self._layer_dense, inputs, **kwargs)
        return self._layer_dense(inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ActivationEmbedding')
class ActivationEmbedding(GraphBaseLayer):

    def __init__(self,
                 activation,
                 activity_regularizer=None,
                 **kwargs):
        """Initialize layer."""
        super(ActivationEmbedding, self).__init__(**kwargs)
        self._layer_act = tf.keras.layers.Activation(activation=activation, activity_regularizer=activity_regularizer)
        self._add_layer_config_to_self = {"_layer_act": ["activation", "activity_regularizer"]}

    def call(self, inputs, **kwargs):
        """Forward pass corresponding to keras Activation layer."""
        return self.call_on_values_tensor_of_ragged(self._layer_act, inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LazyAdd')
class LazyAdd(GraphBaseLayer):

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(LazyAdd, self).__init__(**kwargs)
        self._layer_add = ks.layers.Add()

    def call(self, inputs, **kwargs):
        """Forward pass corresponding to keras Add layer."""
        return self.call_on_values_tensor_of_ragged(self._layer_add, inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LazySubtract')
class LazySubtract(GraphBaseLayer):

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(LazySubtract, self).__init__(**kwargs)
        self._layer_subtract = ks.layers.Subtract()

    def call(self, inputs, **kwargs):
        """Forward pass corresponding layer."""
        return self.call_on_values_tensor_of_ragged(self._layer_subtract, inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LazyAverage')
class LazyAverage(GraphBaseLayer):

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(LazyAverage, self).__init__(**kwargs)
        self._layer_avg = ks.layers.Average()

    def call(self, inputs, **kwargs):
        """Forward pass corresponding to keras Average layer."""
        return self.call_on_values_tensor_of_ragged(self._layer_avg, inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LazyMultiply')
class LazyMultiply(GraphBaseLayer):

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(LazyMultiply, self).__init__(**kwargs)
        self._layer_mult = ks.layers.Multiply()

    def call(self, inputs, **kwargs):
        """Forward pass corresponding to keras Multiply layer."""
        return self.call_on_values_tensor_of_ragged(self._layer_mult, inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='DropoutEmbedding')
class DropoutEmbedding(GraphBaseLayer):

    def __init__(self,
                 rate,
                 noise_shape=None,
                 seed=None,
                 **kwargs):
        """Initialize layer same as Activation."""
        super(DropoutEmbedding, self).__init__(**kwargs)
        self._layer_drop = ks.layers.Dropout(rate=rate, noise_shape=noise_shape, seed=seed)
        self._add_layer_config_to_self = {"_layer_drop": ["rate", "noise_shape", "seed"]}

    def call(self, inputs, **kwargs):
        """Forward pass corresponding to keras Dropout layer."""
        return self.call_on_values_tensor_of_ragged(self._layer_drop, inputs, **kwargs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='LazyConcatenate')
class LazyConcatenate(GraphBaseLayer):

    def __init__(self,
                 axis=-1,
                 **kwargs):
        """Initialize layer."""
        super(LazyConcatenate, self).__init__(**kwargs)
        self.axis = axis

    def build(self, input_shape):
        """Build layer from input shape."""
        super(LazyConcatenate, self).build(input_shape)
        if not isinstance(input_shape, (tuple, list)) or len(input_shape) < 1:
            raise ValueError(
                'A `Concatenate` layer should be called on a list of '
                f'at least 1 input. Received: input_shape={input_shape}')
        if all(shape is None for shape in input_shape):
            return
        # Make sure all the shapes have same ranks.
        ranks = set(len(shape) for shape in input_shape)
        if len(ranks) == 1:
            # Make axis positive then for call on values.
            self.axis = get_positive_axis(self.axis, len(input_shape[0]))

    def call(self, inputs, **kwargs):
        """Forward pass. Concatenate possibly ragged tensors.

        Args:
            inputs (list): List of tensors to concatenate.

        Returns:
            tf.tensor: Single concatenated tensor.
        """
        return self.call_on_values_tensor_of_ragged(tf.concat, inputs, axis=self.axis)

    def get_config(self):
        config = super(LazyConcatenate, self).get_config()
        config.update({"axis": self.axis})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ExpandDims')
class ExpandDims(GraphBaseLayer):

    def __init__(self,
                 axis=-1,
                 **kwargs):
        """Initialize layer."""
        super(ExpandDims, self).__init__(**kwargs)
        self.axis = axis

    def build(self, input_shape):
        super(ExpandDims, self).build(input_shape)
        # If rank is not defined can't call on values, if axis does not happen to be positive.
        if len(input_shape) == 0:
            return
        # The possible target axis can be one rank larger to increase rank with expand_dims.
        self.axis = get_positive_axis(self.axis, len(input_shape) + 1)

    def call(self, inputs, **kwargs):
        """Forward pass wrapping tf.keras layer."""
        return self.call_on_values_tensor_of_ragged(tf.expand_dims, inputs, axis=self.axis)

    def get_config(self):
        config = super(ExpandDims, self).get_config()
        config.update({"axis": self.axis})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ZerosLike')
class ZerosLike(GraphBaseLayer):
    """Make a zero-like graph tensor. Calls tf.zeros_like()."""

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(ZerosLike, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(ZerosLike, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Tensor of node or edge embeddings of shape (batch, [N], F)

        Returns:
            tf.RaggedTensor: Zero-like tensor of input.
        """
        return self.call_on_values_tensor_of_ragged(tf.zeros_like, inputs)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ReduceSum')
class ReduceSum(GraphBaseLayer):
    """Make a zero-like graph tensor. Calls tf.zeros_like()."""

    def __init__(self, axis=-1, **kwargs):
        """Initialize layer."""
        super(ReduceSum, self).__init__(**kwargs)
        self.axis = axis

    def build(self, input_shape):
        """Build layer."""
        super(ReduceSum, self).build(input_shape)
        # If rank is not defined can't call on values, if axis does not happen to be positive.
        if len(input_shape) == 0:
            return
        # Set axis to be positive for defined rank to call on values.
        self.axis = get_positive_axis(self.axis, len(input_shape))

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Tensor of node or edge embeddings of shape (batch, [N], F)

        Returns:
            tf.RaggedTensor: Zero-like tensor of input.
        """
        return self.call_on_values_tensor_of_ragged(tf.reduce_sum, inputs, axis=self.axis)

    def get_config(self):
        config = super(ReduceSum, self).get_config()
        config.update({"axis": self.axis})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='OptionalInputEmbedding')
class OptionalInputEmbedding(GraphBaseLayer):
    """Optional Embedding layer."""

    def __init__(self,
                 input_dim,
                 output_dim,
                 use_embedding=False,
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 activity_regularizer=None,
                 embeddings_constraint=None,
                 mask_zero=False,
                 input_length=None,
                 **kwargs):
        """Initialize layer."""
        super(OptionalInputEmbedding, self).__init__(**kwargs)
        self.use_embedding = use_embedding

        if use_embedding:
            self._layer_embed = ks.layers.Embedding(input_dim=input_dim, output_dim=output_dim,
                                                    embeddings_initializer=embeddings_initializer,
                                                    embeddings_regularizer=embeddings_regularizer,
                                                    activity_regularizer=activity_regularizer,
                                                    embeddings_constraint=embeddings_constraint,
                                                    mask_zero=mask_zero, input_length=input_length)
            self._add_layer_config_to_self = {"_layer_embed": ["input_dim", "output_dim", "embeddings_initializer",
                                                               "embeddings_regularizer", "activity_regularizer",
                                                               "embeddings_constraint", "mask_zero", "input_length"]}

    def build(self, input_shape):
        """Build layer."""
        super(OptionalInputEmbedding, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Tensor of number or embeddings of shape (batch, [N]) or (batch, [N], F)

        Returns:
            tf.RaggedTensor: Zero-like tensor of input.
        """
        if self.use_embedding:
            return self._layer_embed(inputs)
        return inputs

    def get_config(self):
        config = super(OptionalInputEmbedding, self).get_config()
        config.update({"use_embedding": self.use_embedding})
        return config



import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.ops.axis import get_positive_axis

from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.modules import LazyAdd, LazyMultiply, DenseEmbedding, LazyConcatenate, ExpandDims
from kgcnn.layers.geom import EuclideanNorm, ScalarProduct
from kgcnn.layers.gather import GatherNodesOutgoing


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PAiNNconv')
class PAiNNconv(GraphBaseLayer):
    """Continuous filter convolution block of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`_ .

    Args:
        units (int): Units for Dense layer.
        conv_pool (str): Pooling method. Default is 'sum'.
        use_bias (bool): Use bias. Default is True.
        activation (str): Activation function. Default is 'kgcnn>shifted_softplus'.
        kernel_regularizer: Kernel regularization. Default is None.
        bias_regularizer: Bias regularization. Default is None.
        activity_regularizer: Activity regularization. Default is None.
        kernel_constraint: Kernel constrains. Default is None.
        bias_constraint: Bias constrains. Default is None.
        kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
        bias_initializer: Initializer for bias. Default is 'zeros'.
    """

    def __init__(self, units,
                 conv_pool='sum',
                 use_bias=True,
                 activation='swish',
                 cutoff=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize Layer."""
        super(PAiNNconv, self).__init__(**kwargs)
        self.conv_pool = conv_pool
        self.units = units
        self.use_bias = use_bias
        self.cutoff = cutoff

        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}
        # Layer
        self.lay_dense1 = DenseEmbedding(units=self.units, activation=activation, use_bias=self.use_bias, **kernel_args)
        self.lay_phi = DenseEmbedding(units=self.units * 3, activation='linear', use_bias=self.use_bias, **kernel_args)
        self.lay_w = DenseEmbedding(units=self.units * 3, activation='linear', use_bias=self.use_bias, **kernel_args)

        self.lay_split = SplitEmbedding(3, axis=-1)
        self.lay_sum = PoolingLocalEdges(pooling_method=conv_pool)
        self.lay_sum_v = PoolingLocalEdges(pooling_method=conv_pool)
        self.gather_n = GatherNodesOutgoing()
        self.gather_v = GatherNodesOutgoing()
        self.lay_mult = LazyMultiply()
        if self.cutoff is not None:
            self.lay_mult_cutoff = LazyMultiply()
        self.lay_exp_vv = ExpandDims(axis=-2)
        self.lay_exp_vw = ExpandDims(axis=-2)
        self.lay_exp_r = ExpandDims(axis=-1)
        self.lay_mult_vv = LazyMultiply()
        self.lay_mult_vw = LazyMultiply()

        self.lay_add = LazyAdd()

    def build(self, input_shape):
        """Build layer."""
        super(PAiNNconv, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass: Calculate edge update.

        Args:
            inputs: [nodes, equivariant, rbf, envelope, r_ij, edge_index]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - equivariant (tf.RaggedTensor): Equivariant node embedding of shape (batch, [N], 3, F)
                - rdf (tf.RaggedTensor): Radial basis expansion pair-wise distance of shape (batch, [M], #Basis)
                - envelope (tf.RaggedTensor): Distance envelope of shape (batch, [N], 1)
                - r_ij (tf.RaggedTensor): Normalized pair-wise distance of shape (batch, [M], 3)
                - edge_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)

        Returns:
            tuple: [ds, dv]

                - ds (tf.RaggedTensor) Updated node features of shape (batch, [N], F)
                - dv (tf.RaggedTensor) Updated equivariant features of shape (batch, [N], F, 3)
        """
        node, equivariant, rbf, envelope, r_ij, indexlist = inputs
        s = self.lay_dense1(node)
        s = self.lay_phi(s)
        s = self.gather_n([s, indexlist])
        w = self.lay_w(rbf)
        if self.cutoff is not None:
            w = self.lay_mult_cutoff([w, envelope])
        sw = self.lay_mult([s, w])
        sw1, sw2, sw3 = self.lay_split(sw)
        ds = self.lay_sum([node, sw1, indexlist])
        vj = self.gather_v([equivariant, indexlist])
        sw2 = self.lay_exp_vv(sw2)
        dv1 = self.lay_mult_vv([sw2, vj])
        sw3 = self.lay_exp_vw(sw3)
        r_ij = self.lay_exp_r(r_ij)
        dv2 = self.lay_mult_vw([sw3, r_ij])
        dv = self.lay_add([dv1, dv2])
        dv = self.lay_sum_v([node, dv, indexlist])
        return ds, dv

    def get_config(self):
        """Update layer config."""
        config = super(PAiNNconv, self).get_config()
        config.update({"conv_pool": self.conv_pool, "units": self.units, "cutoff": self.cutoff})
        config_dense = self.lay_dense1.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            config.update({x: config_dense[x]})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PAiNNUpdate')
class PAiNNUpdate(GraphBaseLayer):
    """Continuous filter convolution of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`_ .

    Args:
        units (int): Units for Dense layer.
        conv_pool (str): Pooling method. Default is 'sum'.
        use_bias (bool): Use bias. Default is True.
        activation (str): Activation function. Default is 'kgcnn>shifted_softplus'.
        kernel_regularizer: Kernel regularization. Default is None.
        bias_regularizer: Bias regularization. Default is None.
        activity_regularizer: Activity regularization. Default is None.
        kernel_constraint: Kernel constrains. Default is None.
        bias_constraint: Bias constrains. Default is None.
        kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
        bias_initializer: Initializer for bias. Default is 'zeros'.
    """

    def __init__(self, units,
                 use_bias=True,
                 activation='swish',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize Layer."""
        super(PAiNNUpdate, self).__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias

        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}
        # Layer
        self.lay_dense1 = DenseEmbedding(units=self.units, activation=activation, use_bias=self.use_bias, **kernel_args)
        self.lay_lin_u = DenseEmbedding(self.units, activation='linear', use_bias=False, **kernel_args)
        self.lay_lin_v = DenseEmbedding(self.units, activation='linear', use_bias=False, **kernel_args)
        self.lay_a = DenseEmbedding(units=self.units * 3, activation='linear', use_bias=self.use_bias, **kernel_args)

        self.lay_scalar_prod = ScalarProduct(axis=2)
        self.lay_norm = EuclideanNorm(axis=2)
        self.lay_concat = LazyConcatenate(axis=-1)
        self.lay_split = SplitEmbedding(3, axis=-1)

        self.lay_mult = LazyMultiply()
        self.lay_exp_v = ExpandDims(axis=-2)
        self.lay_mult_vv = LazyMultiply()
        self.lay_add = LazyAdd()

    def build(self, input_shape):
        """Build layer."""
        super(PAiNNUpdate, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass: Calculate edge update.

        Args:
            inputs: [nodes, equivariant]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - equivariant (tf.RaggedTensor): Equivariant node embedding of shape (batch, [N], 3, F)

        Returns:
            tuple: [ds, dv]

                - ds (tf.RaggedTensor) Updated node features of shape (batch, [N], F)
                - dv (tf.RaggedTensor) Updated equivariant features of shape (batch, [N], 3, F)
        """
        node, equivariant = inputs
        v_v = self.lay_lin_v(equivariant, **kwargs)
        v_u = self.lay_lin_u(equivariant, **kwargs)
        v_prod = self.lay_scalar_prod([v_u, v_v], **kwargs)
        v_norm = self.lay_norm(v_v, **kwargs)
        a = self.lay_concat([node, v_norm], **kwargs)
        a = self.lay_dense1(a, **kwargs)
        a = self.lay_a(a, **kwargs)
        a_vv, a_sv, a_ss = self.lay_split(a, **kwargs)
        a_vv = self.lay_exp_v(a_vv, **kwargs)
        dv = self.lay_mult_vv([a_vv, v_u], **kwargs)
        ds = self.lay_mult([v_prod, a_sv], **kwargs)
        ds = self.lay_add([ds, a_ss], **kwargs)
        return ds, dv

    def get_config(self):
        """Update layer config."""
        config = super(PAiNNUpdate, self).get_config()
        config.update({"units": self.units})
        config_dense = self.lay_dense1.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            config.update({x: config_dense[x]})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EquivariantInitialize')
class EquivariantInitialize(GraphBaseLayer):
    """Zero equivariant initializer of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`_ .

    Args:
        dim (int): Dimension of equivariant features. Default is 3.
    """

    def __init__(self, dim=3, **kwargs):
        """Initialize Layer."""
        super(EquivariantInitialize, self).__init__(**kwargs)
        self.dim = int(dim)

    def build(self, input_shape):
        """Build layer."""
        super(EquivariantInitialize, self).build(input_shape)
        assert len(input_shape) >= 3, "ERROR:kgcnn: Need input shape of form (batch, None, F_dim)."

    def call(self, inputs, **kwargs):
        """Forward pass: Calculate edge update.

        Args:
            inputs: nodes

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)

        Returns:
            tf.RaggedTensor: Zero equivariant tensor of shape (batch, [N], dim, F)
        """
        self.assert_ragged_input_rank(inputs)
        values = tf.zeros_like(inputs.values)
        values = tf.expand_dims(values, axis=1)
        # Static shape expansion for dim, tf.repeat would be possible too.
        shape_expand = [1] * len(values.shape)
        shape_expand[1] = self.dim
        values = values * tf.zeros(shape_expand)
        out = tf.RaggedTensor.from_row_splits(values, inputs.row_splits)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(EquivariantInitialize, self).get_config()
        config.update({"dim": self.dim})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='SplitEmbedding')
class SplitEmbedding(GraphBaseLayer):
    """Split embeddings of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`_ .

    Args:
        num_or_size_splits: Number or size of splits.
        axis (int): Axis to split.
        num (int): Number of output splits.
    """

    def __init__(self,
                 num_or_size_splits,
                 axis=-1,
                 num=None,
                 **kwargs):
        super(SplitEmbedding, self).__init__(**kwargs)
        self.num_or_size_splits = num_or_size_splits
        self.axis = axis
        self.out_num = num

    def build(self, input_shape):
        super(SplitEmbedding, self).build(input_shape)
        # If rank is not defined can't call on values if axis does not happen to be positive.
        self.axis = get_positive_axis(self.axis, len(input_shape))
        if self.axis <= 1:
            raise ValueError("Can not split tensor at axis <= 1.")

    def call(self, inputs, **kwargs):
        r"""Forward pass: Split embeddings across feature dimension e.g. `axis=-1`..

        Args:
            inputs (tf.RaggedTensor): Embeddings of shape (batch, [N], F)

        Returns:
            list: List of tensor splits of shape (batch, [N], F/num)
        """
        self.assert_ragged_input_rank(inputs, ragged_rank=1)
        # Axis will be positive and >=1 from built!
        # Axis for values is axis-1.
        out_tensors = tf.split(inputs.values, self.num_or_size_splits, axis=self.axis-1, num=self.out_num)
        return [tf.RaggedTensor.from_row_splits(x, inputs.row_splits, validate=self.ragged_validate) for x in
                out_tensors]

    def get_config(self):
        config = super(SplitEmbedding, self).get_config()
        config.update({"num_or_size_splits": self.num_or_size_splits, "axis": self.axis, "num": self.out_num})
        return config



import os
import rdkit
import rdkit.Chem
import rdkit.Chem.AllChem
from openbabel import openbabel
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

# from openbabel import pybel


def single_smile_to_mol(smile: str,
                        sanitize: bool = True,
                        add_hydrogen: bool = True,
                        make_conformers: bool = True,
                        optimize_conformer: bool = True):
    """Workflow to make mol information, i.e. compute structure and conformation. With rdkit and openbabel only.

    Args:
        smile (str): List of smiles.

    Returns:
        list: list with mol strings.
    """
    # Try rdkit first
    try:
        m = rdkit.Chem.MolFromSmiles(smile)
        if sanitize:
            rdkit.Chem.SanitizeMol(m)
        if add_hydrogen:
            m = rdkit.Chem.AddHs(m)  # add H's to the molecule
        m.SetProp("_Name", smile)
        if make_conformers:
            rdkit.Chem.RemoveStereochemistry(m)
            rdkit.Chem.AssignStereochemistry(m)
            rdkit.Chem.AllChem.EmbedMolecule(m, useRandomCoords=True)
        if optimize_conformer and make_conformers:
            rdkit.Chem.AllChem.MMFFOptimizeMolecule(m)
            rdkit.Chem.AssignAtomChiralTagsFromStructure(m)
            rdkit.Chem.AssignStereochemistryFrom3D(m)
            rdkit.Chem.AssignStereochemistry(m)
    except:
        m = None

    if m is not None:
        return rdkit.Chem.MolToMolBlock(m)

    # Try openbabel next
    try:
        m = openbabel.OBMol()
        obconversion = openbabel.OBConversion()
        format_okay = obconversion.SetInAndOutFormats("smi", "mol")
        read_okay = obconversion.ReadString(m, smile)
        is_okay = {"format_okay": format_okay, "read_okay": read_okay}
        if make_conformers:
            # We need to make conformer with builder
            builder = openbabel.OBBuilder()
            build_okay = builder.Build(m)
            is_okay.update({"build_okay": build_okay})
        if add_hydrogen:
            # it seems h's are made after build, an get embedded too
            m.AddHydrogens()
        if optimize_conformer and make_conformers:
            ff = openbabel.OBForceField.FindType("mmff94")
            ff_setup_okay = ff.Setup(m)
            ff.SteepestDescent(100)  # defaults are 50-500 in pybel
            ff.GetCoordinates(m)
            is_okay.update({"ff_setup_okay": ff_setup_okay})
        all_okay = all(list(is_okay.values()))
        if not all_okay:
            print("WARNING: Openbabel returned false flag %s" % [key for key, value in is_okay.items() if not value])
    except:
        m = None
        obconversion = None

    if m is not None:
        return obconversion.WriteString(m)

    print("WARNING: Openbabel and RDKit failed for %s." % smile)
    return None


def smile_to_mol_parallel(smile_list: list,
                          num_workers: int = None,
                          sanitize: bool = True,
                          add_hydrogen: bool = True,
                          make_conformers: bool = True,
                          optimize_conformer: bool = True):
    if num_workers is None:
        num_workers = os.cpu_count()

    # Default parallelize with rdkit and openbabel
    if num_workers == 1:
        mol_list = [single_smile_to_mol(x, sanitize=sanitize, add_hydrogen=add_hydrogen,
                                        make_conformers=make_conformers,
                                        optimize_conformer=optimize_conformer) for x in smile_list]
        return mol_list
    else:
        arg_list = [(x, sanitize, add_hydrogen, make_conformers, optimize_conformer) for x in smile_list]
        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            result = executor.map(single_smile_to_mol, *zip(*arg_list))
        mol_list = list(result)
        return mol_list


# # Former parallelization using multiprocessing
# if __name__ == '__main__':
#     # Input arguments from command line.
#     import multiprocessing as mp
#     import argparse
#     from distutils import util
#     from kgcnn.mol.io import read_smiles_file, write_mol_block_list_to_sdf
#     parser = argparse.ArgumentParser(description='Translate smiles.')
#     parser.add_argument("--file", required=True, help="Path to smile file")
#     parser.add_argument("--nprocs", required=False, help="nprocs", default=1, type=int)
#     parser.add_argument("--sanitize", required=False, help="sanitize", default=True, type=util.strtobool)
#     parser.add_argument("--add_hydrogen", required=False, help="add_hydrogen", default=True, type=util.strtobool)
#     parser.add_argument("--make_conformers", required=False, help="make_conformers", default=True, type=util.strtobool)
#     parser.add_argument("--optimize_conformer", required=False, help="optimize_conformer", default=True, type=util.strtobool)
#     args = vars(parser.parse_args())
#
#     arg_nprocs = args["nprocs"]
#     arg_file_path = args["file"]
#     arg_sanitize = args["sanitize"]
#     arg_add_hydrogen = args["add_hydrogen"]
#     arg_make_conformers = args["make_conformers"]
#     arg_optimize_conformer = args["optimize_conformer"]
#
#     smile_list = read_smiles_file(arg_file_path)
#     if arg_nprocs is None:
#         arg_nprocs = mp.cpu_count()
#
#     pool = mp.Pool(arg_nprocs)
#     results = pool.starmap(single_smile_to_mol,
#                            [(smile, arg_sanitize, arg_add_hydrogen, arg_make_conformers, arg_optimize_conformer)
#                             for smile in smile_list])
#     pool.close()
#
#     write_mol_block_list_to_sdf(results, os.path.splitext(arg_file_path)[0] + ".sdf")



import tensorflow as tf
import tensorflow.keras as ks


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ScaledMeanAbsoluteError')
class ScaledMeanAbsoluteError(tf.keras.metrics.MeanAbsoluteError):
    """Metric for a scaled mean absolute error (MAE), which can undo a pre-scaling of the targets. Only intended as
    metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='mean_absolute_error', **kwargs):
        super(ScaledMeanAbsoluteError, self).__init__(name=name, **kwargs)
        self.scale = self.add_weight(shape=scaling_shape,
                                     initializer=tf.keras.initializers.Ones(), name='kgcnn_scale_mae',
                                     dtype=tf.keras.backend.floatx())
        self.scaling_shape = scaling_shape

    def reset_state(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_mae' not in v.name])
        # Or set them explicitly.
        # ks.backend.set_value(self.total, 0)
        # ks.backend.set_value(self.count, 0)

    def reset_states(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_mae' not in v.name])
        # Or set them explicitly.
        # ks.backend.set_value(self.total, 0)
        # ks.backend.set_value(self.count, 0)

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = self.scale * y_true
        y_pred = self.scale * y_pred
        return super(ScaledMeanAbsoluteError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        mae_conf = super(ScaledMeanAbsoluteError, self).get_config()
        mae_conf.update({"scaling_shape": self.scaling_shape})
        return mae_conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        ks.backend.set_value(self.scale, scale)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ScaledRootMeanSquaredError')
class ScaledRootMeanSquaredError(tf.keras.metrics.RootMeanSquaredError):
    """Metric for a scaled root mean squared error (RMSE), which can undo a pre-scaling of the targets.
    Only intended as metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='root_mean_squared_error', **kwargs):
        super(ScaledRootMeanSquaredError, self).__init__(name=name, **kwargs)
        self.scale = self.add_weight(shape=scaling_shape,
                                     initializer=tf.keras.initializers.Ones(), name='kgcnn_scale_rmse',
                                     dtype=tf.keras.backend.floatx())
        self.scaling_shape = scaling_shape

    def reset_state(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_rmse' not in v.name])
        # Or set them explicitly.
        # ks.backend.set_value(self.total, 0)
        # ks.backend.set_value(self.count, 0)

    def reset_states(self):
        ks.backend.batch_set_value([(v, 0) for v in self.variables if 'kgcnn_scale_rmse' not in v.name])
        # Or set them explicitly.
        # ks.backend.set_value(self.total, 0)
        # ks.backend.set_value(self.count, 0)

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = self.scale * y_true
        y_pred = self.scale * y_pred
        return super(ScaledRootMeanSquaredError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        mae_conf = super(ScaledRootMeanSquaredError, self).get_config()
        mae_conf.update({"scaling_shape": self.scaling_shape})
        return mae_conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        ks.backend.set_value(self.scale, scale)



import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.gather import GatherNodesOutgoing
from kgcnn.layers.modules import DenseEmbedding, LazyConcatenate, ActivationEmbedding, LazyAdd, DropoutEmbedding, \
    OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingLocalEdges, PoolingNodes
from kgcnn.layers.conv.dmpnn_conv import DMPNNPPoolingEdgesDirected
from kgcnn.utils.models import update_model_kwargs

# Analyzing Learned Molecular Representations for Property Prediction
# by Kevin Yang, Kyle Swanson, Wengong Jin, Connor Coley, Philipp Eiden, Hua Gao,
# Angel Guzman-Perez, Timothy Hopper, Brian Kelley, Miriam Mathea, Andrew Palmer,
# Volker Settels, Tommi Jaakkola, Klavs Jensen, and Regina Barzilay
# https://pubs.acs.org/doi/full/10.1021/acs.jcim.9b00237

model_default = {'name': "DMPNN",
                 'inputs': [
                     {'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                     {'shape': (None,), 'name': "edge_attributes", 'dtype': 'float32', 'ragged': True},
                     {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True},
                     {'shape': (None, 1), 'name': "edge_indices_reverse", 'dtype': 'int64', 'ragged': True}],
                 'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                     "edge": {"input_dim": 5, "output_dim": 64}},
                 'pooling_args': {'pooling_method': "sum"},
                 "edge_initialize": {"units": 128, 'use_bias': True, 'activation': 'relu'},
                 'edge_dense': {"units": 128, 'use_bias': True, 'activation': 'linear'},
                 "edge_activation": {"activation": "relu"},
                 "node_dense": {"units": 128, 'use_bias': True, 'activation': 'relu'},
                 'verbose': 10, "depth": 5, "dropout": {"rate": 0.1},
                 'output_embedding': 'graph',
                 'output_mlp': {"use_bias": [True, True, False], "units": [64, 32, 1],
                                "activation": ['relu', 'relu', 'linear']}
                 }


@update_model_kwargs(model_default)
def make_model(name=None,
               inputs=None,
               input_embedding=None,
               pooling_args=None,
               edge_initialize=None,
               edge_dense=None,
               edge_activation=None,
               node_dense=None,
               dropout=None,
               depth=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make DMPNN graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        name (str): Name of the model. Should be "DMPNN".
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        pooling_args (dict): Dictionary of layer arguments unpacked in `PoolingNodes`, `PoolingLocalEdges` layers.
        edge_initialize (dict): Dictionary of layer arguments unpacked in `Dense` layer for first edge embedding.
        edge_dense (dict): Dictionary of layer arguments unpacked in `Dense` layer for edge embedding.
        edge_activation (dict): Edge Activation after skip connection.
        node_dense (dict): Dense kwargs for node embedding layer.
        depth (int): Number of graph embedding units or depth of the network.
        dropout (dict): Dictionary of layer arguments unpacked in `Dropout`.
        verbose (int): Level for print information.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    edge_pair_input = ks.layers.Input(**inputs[3])
    ed_pairs = edge_pair_input

    # Embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    ed = OptionalInputEmbedding(**input_embedding['edge'],
                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
    edi = edge_index_input

    # Make first edge hidden h0
    h_n0 = GatherNodesOutgoing()([n, edi])
    h0 = LazyConcatenate(axis=-1)([h_n0, ed])
    h0 = DenseEmbedding(**edge_initialize)(h0)

    # One Dense layer for all message steps
    edge_dense_all = DenseEmbedding(**edge_dense)  # Should be linear activation

    # Model Loop
    h = h0
    for i in range(depth):
        m_vw = DMPNNPPoolingEdgesDirected()([n, h, edi, ed_pairs])
        h = edge_dense_all(m_vw)
        h = LazyAdd()([h, h0])
        h = ActivationEmbedding(**edge_activation)(h)
        if dropout is not None:
            h = DropoutEmbedding(**dropout)(h)

    mv = PoolingLocalEdges(**pooling_args)([n, h, edi])
    mv = LazyConcatenate(axis=-1)([mv, n])
    hv = DenseEmbedding(**node_dense)(mv)

    # Output embedding choice
    n = hv
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_args)(n)
        # final dense layers
        main_output = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        main_output = GraphMLP(**output_mlp)(out)
        # no ragged for distribution supported atm
        main_output = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(main_output)
    else:
        raise ValueError("Unsupported graph embedding for mode `DMPNN`")

    model = ks.models.Model(inputs=[node_input, edge_input, edge_index_input, edge_pair_input], outputs=main_output)
    return model



import tensorflow as tf


@tf.function
def segment_softmax(data, segment_ids, normalize: bool = True):
    """Segment softmax similar to segment_max but with a softmax function.

    Args:
        data (tf.Tensor): Data tensor that has sorted segments.
        segment_ids (tf.Tensor): IDs of the segments.
        normalize (bool): Normalize data for softmax. Default is True.

    Returns:
        tf.Tensor: reduced segment data with a softmax function.
    """
    if normalize:
        data_segment_max = tf.math.segment_max(data, segment_ids)
        data_max = tf.gather(data_segment_max, segment_ids)
        data = data - data_max

    data_exp = tf.math.exp(data)
    data_exp_segment_sum = tf.math.segment_sum(data_exp, segment_ids)
    data_exp_sum = tf.gather(data_exp_segment_sum, segment_ids)
    return data_exp / data_exp_sum


@tf.function
def segment_ops_by_name(segment_name: str, data, segment_ids):
    """Segment operation chosen by string identifier.

    Args:
        segment_name (str): Name of the segment operation.
        data (tf.Tensor): Data tensor that has sorted segments.
        segment_ids (tf.Tensor): IDs of the segments.

    Returns:
        tf.Tensor: reduced segment data with method by segment_name.
    """
    if segment_name in ["segment_mean", "mean", "reduce_mean"]:
        pool = tf.math.segment_mean(data, segment_ids)
    elif segment_name in ["segment_sum", "sum", "reduce_sum"]:
        pool = tf.math.segment_sum(data, segment_ids)
    elif segment_name in ["segment_max", "max", "reduce_max"]:
        pool = tf.math.segment_max(data, segment_ids)
    elif segment_name in ["segment_min", "sum", "reduce_min"]:
        pool = tf.math.segment_min(data, segment_ids)
    elif segment_name in ["segment_softmax", "segment_soft_max", "softmax", "soft_max", "reduce_softmax"]:
        pool = segment_softmax(data, segment_ids)
    else:
        raise TypeError("Unknown segment operation, choose: 'segment_mean', 'segment_sum', ...")
    return pool



import numpy as np
import tensorflow as tf

from kgcnn.utils.data import ragged_tensor_from_nested_numpy


class NumpyTensorList:
    r"""An utility class to keep lists of arrays. Multiple lists are handled in parallel and therefore should
    have the same size. The behavior should be similar to :obj:`train_test_split` from scikit-learn.
    The length of each list should match the length of the dataset and
    each list-item be e.g. a numpy-tensor for a data-point. The lists could be X, y or different graph attributes.
    This class is intentionally kept simple but can be built on in the future.

    """

    def __init__(self, *args):
        """Make a new `NumpyTensorList` object with list of matching data like features and labels of a dataset.

        .. code-block:: python

            import numpy as np
            data = NumpyTensorList([np.array([0,0,0]), np.array([0])], [np.array([1]), np.array([1,1])])
            print(data[0].tensor())

        Args:
            args: Multiple lists of np.ndarrays.
        """
        self._tensor_list = [x for x in args]
        if len(self._tensor_list) <= 0:
            print("Received empty list input for `NumpyTensorList`. \
                   Expected one or more list of numpy arrays.")

        # check that all have same length
        self._length_test()

    def _length_test(self):
        length_test = np.array([len(x) for x in self._tensor_list], dtype="int")
        if len(length_test) > 0:
            if not np.all(length_test == length_test[0]):
                print("Length of list input to `NumpyTensorList` are different.")

    def __getitem__(self, item):
        r"""Indexing or getitem method to apply to each list and return a new :obj:`NumpyTensorList`.

        Args:
            item (list, int, slice, np.ndarray): Index or list of indices to collect from each list in self.

        Returns:
            NumpyTensorList: New :obj:`NumpyTensorList` with only `items` in each list.
        """
        if isinstance(item, int):
            return NumpyTensorList(*[[x[item]] for x in self._tensor_list])
        elif isinstance(item, slice):
            return NumpyTensorList(*[x[item] for x in self._tensor_list])
        elif hasattr(item, "__getitem__"):
            return NumpyTensorList(*[[x[y] for y in item] for x in self._tensor_list])
        else:
            raise ValueError("ERROR:kgcnn: Can not iterate over input of getitem() for `NumpyTensorList`: ", item)

    def tensor(self, ragged=None):
        """Cast list of numpy arrays to tf.Tensor or tf.RaggedTensor objects.

        Note: A copy of the data may be generated for ragged tensors!

        Args:
            ragged (list): Information whether a output tensor for corresponding list is ragged. Default is None.

        Returns:
            list: A list of tf.Tensor or tf.RaggedTensor objects.
        """
        # TODO: We can check if tf.RaggedTensor is actually necessary, without relying on ragged-argument.
        # TODO: Also need to improve casting, e.g. memory usage
        # TODO: Also ragged_tensor_from_nested_numpy will fail for arbitrary ragged shape
        if ragged is None:
            ragged = [False] * len(self._tensor_list)
        assert len(ragged) == len(self._tensor_list)
        out_list = []
        for i, x in enumerate(self._tensor_list):
            if not ragged[i]:
                out_list.append(tf.constant(x))
            else:
                out_list.append(ragged_tensor_from_nested_numpy(x))
        return out_list

    def pop(self, index: int):
        r"""Remove a single item at index from each list within `NumpyTensorList`.

        Args:
            index (int): Index of item to remove.

        Returns:
            list: A list of removed items. Number of lists stored by `NumpyTensorList`.
        """
        removed_items = []
        for i, x in enumerate(self._tensor_list):
            if isinstance(x, list):
                temp = x.pop(index)
                removed_items.append(temp)
            else:
                removed_items.append(x[index])
                self._tensor_list[i] = [x[j] for j in range(len(x)) if j != index]
        return removed_items

    def __len__(self):
        """Get length of each list."""
        self._length_test()
        return len(self._tensor_list[0])




import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.gather import GatherNodesOutgoing
from kgcnn.layers.modules import LazyConcatenate, OptionalInputEmbedding
from kgcnn.layers.norm import GraphLayerNormalization
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes, PoolingLocalMessages, PoolingLocalEdgesLSTM
from kgcnn.utils.models import update_model_kwargs

# 'Inductive Representation Learning on Large Graphs'
# William L. Hamilton and Rex Ying and Jure Leskovec
# http://arxiv.org/abs/1706.02216


hyper_model_default = {'name': "GraphSAGE",
                       'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                                  {'shape': (None,), 'name': "edge_attributes", 'dtype': 'float32', 'ragged': True},
                                  {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                       'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                           "edge": {"input_dim": 5, "output_dim": 64}},
                       'node_mlp_args': {"units": [100, 50], "use_bias": True, "activation": ['relu', "linear"]},
                       'edge_mlp_args': {"units": [100, 50], "use_bias": True, "activation": ['relu', "linear"]},
                       'pooling_args': {'pooling_method': "segment_mean"}, 'gather_args': {},
                       'concat_args': {"axis": -1},
                       'use_edge_features': True, 'pooling_nodes_args': {'pooling_method': "mean"},
                       'depth': 3, 'verbose': 10,
                       'output_embedding': 'graph',
                       'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
                                      "activation": ['relu', 'relu', 'sigmoid']}
                       }


@update_model_kwargs(hyper_model_default)
def make_model(inputs=None,
               input_embedding=None,
               node_mlp_args=None,
               edge_mlp_args=None,
               pooling_args=None,
               pooling_nodes_args=None,
               gather_args=None,
               concat_args=None,
               use_edge_features=None,
               depth=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make GraphSAGE graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        node_mlp_args (dict): Dictionary of layer arguments unpacked in `MLP` layer for node updates.
        edge_mlp_args (dict): Dictionary of layer arguments unpacked in `MLP` layer for edge updates.
        pooling_args (dict): Dictionary of layer arguments unpacked in `PoolingLocalMessages` layer.
        pooling_nodes_args (dict): Dictionary of layer arguments unpacked in `PoolingNodes` layer.
        gather_args (dict): Dictionary of layer arguments unpacked in `GatherNodes` layer.
        concat_args (dict): Dictionary of layer arguments unpacked in `LazyConcatenate` layer.
        use_edge_features (bool): Whether to add edge features in message step.
        depth (int): Number of graph embedding units or depth of the network.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input embedding, if no feature dimension
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    ed = OptionalInputEmbedding(**input_embedding['edge'],
                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
    edi = edge_index_input

    for i in range(0, depth):
        # upd = GatherNodes()([n,edi])
        eu = GatherNodesOutgoing(**gather_args)([n, edi])
        if use_edge_features:
            eu = LazyConcatenate(**concat_args)([eu, ed])

        eu = GraphMLP(**edge_mlp_args)(eu)
        # Pool message
        if pooling_args['pooling_method'] in ["LSTM", "lstm"]:
            nu = PoolingLocalEdgesLSTM(**pooling_args)([n, eu, edi])
        else:
            nu = PoolingLocalMessages(**pooling_args)([n, eu, edi])  # Summing for each node connection

        nu = LazyConcatenate(**concat_args)([n, nu])  # LazyConcatenate node features with new edge updates

        n = GraphMLP(**node_mlp_args)(nu)
        n = GraphLayerNormalization()(n)  # Normalize

    # Regression layer on output
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_nodes_args)(n)
        out = ks.layers.Flatten()(out)  # will be tensor
        main_output = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)(n)
        main_output = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported graph embedding for mode `GraphSAGE`")

    model = tf.keras.models.Model(inputs=[node_input, edge_input, edge_index_input], outputs=main_output)
    return model
